# This file is part of sympy2c.
#
# Copyright (C) 2013-2022 ETH Zurich, Institute for Particle and Astrophysics and SIS
# ID.
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.

import hashlib
from collections import OrderedDict

import sympy as sp

from .globals import Globals
from .integral import IntegralFunctionWrapper


class Sp2CVisitor(object):
    def __init__(self, globals_=None):
        self.extra_wrappers = OrderedDict()
        if globals_ is None:
            globals_ = Globals()
        self.globals_ = globals_
        self.integral_count = 0

    def visit(self, expr):
        if hasattr(self, "visit_{0}".format(type(expr).__name__)):
            return getattr(self, "visit_{0}".format(type(expr).__name__))(expr)
        else:
            return self.generic_visit(expr)

    def generic_visit(self, expr):
        if hasattr(expr, "to_c"):
            return expr.to_c(self)
        raise Exception(
            "Not Implemented Expression: {0}: {1!s}".format(type(expr).__name__, expr)
        )

    def visit_Add(self, expr):
        return "({0})".format(" + ".join([self.visit(arg) for arg in expr.args]))

    def visit_Mul(self, expr):
        return " * ".join([self.visit(arg) for arg in expr.args])

    def visit_besselj(self, expr):
        order, arg = expr.args
        return "gsl_sf_bessel_Jn({0}, {1})".format(self.visit(order), self.visit(arg))

    def visit_besseli(self, expr):
        order, arg = expr.args
        return "gsl_sf_bessel_In({0}, {1})".format(self.visit(order), self.visit(arg))

    def visit_besselk(self, expr):
        order, arg = expr.args
        return "gsl_sf_bessel_Kn({0}, {1})".format(self.visit(order), self.visit(arg))

    def visit_bessely(self, expr):
        order, arg = expr.args
        return "gsl_sf_bessel_Yn({0}, {1})".format(self.visit(order), self.visit(arg))

    def visit_gamma(self, expr):
        arg = expr.args[0]
        return "gsl_sf_gamma({0})".format(self.visit(arg))

    def visit_Pow(self, expr):
        if isinstance(expr.exp, sp.Number):
            power = float(expr.exp)
            sign = -1 if power < 0 else 1

            if sign * expr.exp == 0.5:
                ret = "sqrtl({0})".format(self.visit(expr.base))
            elif sign * expr.exp == 1:
                ret = self.visit(expr.base)
            elif sign * expr.exp == 2:
                ret = "{0} * {0}".format(self.visit(expr.base))
            elif sign * expr.exp == 3:
                ret = "{0} * {0} * {0}".format(self.visit(expr.base))
            elif sign * expr.exp == 4:
                ret = "{0} * {0} * {0} * {0}".format(self.visit(expr.base))
            else:
                ret = "pow({0}, {1})".format(self.visit(expr.base), abs(power))
            return "1. / ({0})".format(ret) if sign < 0 else ret

        return "pow({0}, {1})".format(self.visit(expr.base), self.visit(expr.exp))

    def visit_Integer(self, expr):
        return str(expr.p)

    def visit_Abs(self, expr):
        return "fabs({})".format(self.visit(expr.args[0]))

    def visit_Rational(self, expr):
        return "{0} / {1}".format(float(expr.p), float(expr.q))

    def visit_Float(self, expr):
        return str(float(expr))

    def visit_Symbol(self, expr):
        return expr.name

    def visit_NegativeOne(self, expr):
        return "-1"

    def visit_Zero(self, expr):
        return "0"

    def visit_One(self, expr):
        return "1"

    def visit_Half(self, expr):
        return "0.5"

    def visit_float(self, expr):
        return str(expr)

    def visit_int(self, expr):
        return str(expr)

    def visit_sin(self, expr):
        return "sin({})".format(self.visit(expr.args[0]))

    def visit_cos(self, expr):
        return "cos({})".format(self.visit(expr.args[0]))

    def visit_log(self, expr):
        return "log({})".format(self.visit(expr.args[0]))

    def visit_exp(self, expr):
        return "exp({})".format(self.visit(expr.args[0]))

    def visit_tanh(self, expr):
        return "tanh({})".format(self.visit(expr.args[0]))

    def visit_Pi(self, expr):
        return "3.14159265358979323846"

    def visit_Infinity(self, expr):
        return "INFINITY"

    def visit_NegativeInfinity(self, expr):
        return "-INFINITY"

    def visit__Vector(self, expr):
        return str(expr)

    def visit_VectorElement(self, expr):
        return str(expr)

    def visit__Integral(self, integral):
        integrand, integration_variable, low, high, id_ = integral.args[:-1]

        # the variables for the integrand. we include the limits here
        # for nested integrals:

        free_variables_integrand = sorted(
            (integrand.free_symbols) - set(self.globals_.variables), key=str
        )

        variables_integrand = free_variables_integrand + [integration_variable]

        args = ["_arg_{}".format(i) for i in range(len(variables_integrand))]

        subs = {free_var: arg for (free_var, arg) in zip(variables_integrand, args)}

        integrand = integrand.subs(subs)

        key = compute_integral_fingerprint(
            integrand, integration_variable, low, high, id_, free_variables_integrand
        )

        integrand_name = "integrand___{}".format(key)
        integral_function_name = "integral___{}".format(key)

        if integral_function_name not in self.extra_wrappers:
            from .function import InternalFunction, function_wrapper_factory

            integrand_as_function = InternalFunction(
                integrand_name,
                integrand.expression,
                *args,
                # *tuple(v.as_function_arg() for v in variables_integrand),
            )

            integrand_wrapper = function_wrapper_factory(
                integrand_as_function, self.globals_, self
            )
            integrand_wrapper.setup_code_generation()
            self.extra_wrappers[integrand_name] = integrand_wrapper

            # low_expression_in_c = self.visit(low)
            # high_expression_in_c = self.visit(high)

            integral_function_wrapper = IntegralFunctionWrapper(
                integral_function_name,
                args[
                    :-1
                ],  # tuple(v.as_function_arg() for v in free_variables_integrand),
                # low_expression_in_c,
                # high_expression_in_c,
                integrand_name,
                id_,
            )

            integral_function_wrapper.setup_code_generation()
            self.extra_wrappers[integral_function_name] = integral_function_wrapper

        low = self.visit(low)
        high = self.visit(high)
        call = "{}({})".format(
            integral_function_name,
            ", ".join(str(var) for var in free_variables_integrand + [low, high]),
        )
        return call

    def visit_Integrand(self, expr):
        return self.visit(expr.expression)

    def visit_IfThenElse(self, expr):
        cond, v1, v2 = expr.args
        return "(({}) ? ({}) : ({}))".format(
            self.visit(cond), self.visit(v1), self.visit(v2)
        )

    def visit_ERROR(self, expr):
        assert hasattr(expr, "message")
        assert isinstance(expr.message, str)
        return 'set_error_message("{}")'.format(expr.message)

    def visit_StrictLessThan(self, expr):
        return "(({}) < ({}))".format(
            self.visit(expr.args[0]), self.visit(expr.args[1])
        )

    def visit_LessThan(self, expr):
        return "(({}) <= ({}))".format(
            self.visit(expr.args[0]), self.visit(expr.args[1])
        )

    def visit_StrictGreaterThan(self, expr):
        return "(({}) > ({}))".format(
            self.visit(expr.args[0]), self.visit(expr.args[1])
        )

    def visit_GreaterThan(self, expr):
        return "(({}) >= ({}))".format(
            self.visit(expr.args[0]), self.visit(expr.args[1])
        )

    def visit_Equality(self, expr):
        return "(({}) == ({}))".format(
            self.visit(expr.args[0]), self.visit(expr.args[1])
        )

    def visit_Unequality(self, expr):
        return "(({}) != ({}))".format(
            self.visit(expr.args[0]), self.visit(expr.args[1])
        )

    def visit_InterpolationFunction1DInstance(self, expr):
        from .interpolation import InterpolationFunction1DWrapper

        self.extra_wrappers[expr._name] = InterpolationFunction1DWrapper(expr, self)
        return "(__ip_{}({}))".format(expr._name, self.visit(expr.argument))


def compute_integral_fingerprint(
    integrand, integration_variable, low, high, id_, free_variables_integrand
):
    integrand = integrand.subs({integration_variable: "__integration_variable"})
    key = hashlib.md5(
        (str(integrand) + str(id_)).encode("utf-8")  # + str(low) + str(high)
    ).hexdigest()
    return key



# This file is part of PyCosmo, a multipurpose cosmology calculation tool in Python.
#
# Copyright (C) 2013-2022 ETH Zurich, Institute for Particle and Astrophysics and SIS
# ID.
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.


from sympy2c import Module, PythonFunction


def test_python_function(compile_and_load):
    def add(a, b):
        return a + b

    f = PythonFunction(add)
    m = Module()
    m.add(f)

    assert m.get_unique_id() is not None

    w = compile_and_load(m)
    assert w.add(1, 2) == 3



# This file is part of PyCosmo, a multipurpose cosmology calculation tool in Python.
#
# Copyright (C) 2013-2022 ETH Zurich, Institute for Particle and Astrophysics and SIS
# ID.
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.


import numpy as np

from sympy2c import Globals, Module, OdeCombined, symbols
from sympy2c.ode_fast import OdeFast


def test_combined_ode(tmpdir, compile_and_load):
    t, y0, y1, y2, st = symbols("t y0 y1 y2 st")

    ydot0 = -0.04 * y0 + 1e4 * y1 * y2
    ydot2 = 3e7 * y1 * y1
    ydot1 = -ydot0 - ydot2

    lhs = [y2, y1, y0]
    rhs = [ydot2, ydot1, ydot0]

    y0_vec = np.array([0.0, 0.0, 1.0])
    atol = np.array([1e-10, 1e-8, 1e-8])
    rtol = 1e-6

    ode_robertson_1 = OdeFast("robertson_1", t, lhs, rhs, reorder=False)
    ode_robertson_1_reordered = OdeFast(
        "robertson_1_reordered", t, lhs, rhs, reorder=True
    )
    ode_robertson_2 = OdeFast("robertson_2", t, lhs[::-1], rhs[::-1], reorder=False)

    tvec = 0.4 * 10 ** np.arange(0, 6)

    g = Globals(st)
    # f = Function("f", st)

    def switch_time(wrapper, a):
        st = wrapper.globals.st
        print("st = ", st)
        return st

    def switch(wrapper, t, y):
        return y[-1, ::-1]

    def merge(wrapper, tvec_0, t_switch, tvec_1, ys_0, ys_1):
        return np.vstack((ys_0, ys_1[:, ::-1]))

    ode_final = OdeCombined(
        "ode_final",
        ode_robertson_1,
        ode_robertson_2,
        switch_time,
        switch,
        merge,
    )
    ode_final_reordered = OdeCombined(
        "ode_final_reordered",
        ode_robertson_1_reordered,
        ode_robertson_2,
        switch_time,
        switch,
        merge,
    )
    module = Module()
    module.add(g)
    module.add(ode_robertson_1)
    module.add(ode_robertson_1_reordered)
    module.add(ode_robertson_2)
    module.add(ode_final)
    module.add(ode_final_reordered)
    # module.add(f)
    w = compile_and_load(module)

    w.globals.st = 400

    result_0, meta_0 = w.solve_fast_ode_final(y0_vec, tvec, rtol, atol)
    result_1, meta_1 = w.solve_fast_robertson_1(y0_vec, tvec, rtol, atol)
    result_2, meta_2 = w.solve_fast_ode_final_reordered(y0_vec, tvec, rtol, atol)
    assert np.allclose(result_0, result_1)
    assert np.allclose(result_0, result_2)
    assert meta_0.keys() == meta_1.keys()

    print(meta_0)
    print(meta_1)
    print(meta_2)

    w.globals.st = 401

    result_0, meta_0 = w.solve_fast_ode_final(y0_vec, tvec, rtol, atol)
    result_1, meta_1 = w.solve_fast_robertson_1(y0_vec, tvec, rtol, atol)
    assert np.allclose(result_0, result_1)
    assert meta_0.keys() == meta_1.keys()

    w.globals.st = 1e222

    result_0, meta_0 = w.solve_fast_ode_final(y0_vec, tvec, rtol, atol)
    result_1, meta_1 = w.solve_fast_robertson_1(y0_vec, tvec, rtol, atol)

    w.globals.st = -10
    result_2, meta_2 = w.solve_fast_robertson_1(y0_vec, tvec, rtol, atol)

    w.globals.st = tvec[0]
    result_3, meta_3 = w.solve_fast_robertson_1(y0_vec, tvec, rtol, atol)

    w.globals.st = tvec[0] - 1
    result_4, meta_4 = w.solve_fast_robertson_1(y0_vec, tvec, rtol, atol)

    w.globals.st = tvec[-1]
    result_5, meta_5 = w.solve_fast_robertson_1(y0_vec, tvec, rtol, atol)

    w.globals.st = tvec[-1] + 1
    result_6, meta_6 = w.solve_fast_robertson_1(y0_vec, tvec, rtol, atol)

    assert np.allclose(result_0, result_1, atol=0)
    assert np.allclose(result_0, result_2, atol=0)
    assert np.allclose(result_0, result_3, atol=0)
    assert np.allclose(result_0, result_4, atol=0)
    assert np.allclose(result_0, result_5, atol=0)
    assert np.allclose(result_0, result_6, atol=0)

    assert meta_0.keys() == meta_1.keys()
    assert meta_0.keys() == meta_2.keys()
    assert meta_0.keys() == meta_3.keys()
    assert meta_0.keys() == meta_4.keys()
    assert meta_0.keys() == meta_5.keys()
    assert meta_0.keys() == meta_6.keys()


def test_combined_ode_different_dimensions(tmpdir, compile_and_load):
    t, y0, y1, y2, y3, st = symbols("t y0 y1 y2 y3 st")

    ydot0 = -0.04 * y0 + 1e4 * y1 * y2
    ydot2 = 3e7 * y1 * y1
    ydot1 = -ydot0 - ydot2
    ydot3 = y3

    lhs = [y0, y1, y2, y3]
    rhs = [ydot0, ydot1, ydot2, ydot3]

    y0_vec = np.array([1.0, 0.0, 0.0, 0.0])
    atol = 1e-10  # np.array([1e-10, 1e-8, 1e-8])
    rtol = 1e-6

    ode_full = OdeFast("ode_full", t, lhs, rhs, reorder=False)
    ode_3 = OdeFast("ode_3", t, lhs[:3], rhs[:3], reorder=False)

    tvec = 0.4 * 10 ** np.arange(0, 6)

    g = Globals(st)

    def switch_time(wrapper, a):
        st = wrapper.globals.st
        return st

    def switch_1(wrapper, t, y):
        if not y.shape[0]:
            return y[:, :-1]
        return y[-1, :-1]

    def merge_1(wrapper, tvec_0, t_switch, tvec_1, ys_0, ys_1):
        if not ys_0.shape[0]:
            return ys_1
        return np.vstack((ys_0[:, :-1], ys_1))

    ode_final_1 = OdeCombined(
        "ode_final_1",
        ode_full,
        ode_3,
        switch_time,
        switch_1,
        merge_1,
    )

    def switch_2(wrapper, t, y):
        return np.append(y[-1], 0.0)

    def merge_2(wrapper, tvec_0, t_switch, tvec_1, ys_0, ys_1):
        if not ys_1.shape[1]:
            return ys_0
        return np.vstack((ys_0, ys_1[:, :-1]))

    ode_final_2 = OdeCombined(
        "ode_final_2",
        ode_3,
        ode_full,
        switch_time,
        switch_2,
        merge_2,
    )

    module = Module()
    module.add(g)
    module.add(ode_3)
    module.add(ode_full)
    module.add(ode_final_1)
    module.add(ode_final_2)
    w = compile_and_load(module)

    for st in (400, tvec[0], tvec[0] - 1, tvec[-1], tvec[-1] + 1):
        w.globals.st = st

        result_1, meta_1 = w.solve_fast_ode_final_1(y0_vec, tvec, rtol, atol)
        result_2, meta_2 = w.solve_fast_ode_final_2(y0_vec[:-1], tvec, rtol, atol)

        assert np.all(result_1 == result_2)


def test_combined_ode_recompile(tmpdir, compile_and_load):
    t, y1, y2, y3 = symbols("t y1 y2 y3")

    ydot1 = -0.04 * y1 + 1e4 * y2 * y3
    ydot3 = 3e7 * y2 * y2
    ydot2 = -ydot1 - ydot3

    # we change order of original system of odes to enforce permutations during
    # modified lsoda solver:
    ode_1 = OdeFast("robertson_1", t, [y3, y2, y1], [ydot3, ydot2, ydot1])
    ode_2 = OdeFast("robertson_2", t, [y3, y2, y1], [ydot3, ydot2, ydot1])

    def switch_time(wrapper, a):
        return 401

    def switch(wrapper, t, y):
        return y[-1, :]

    def merge(wrapper, tvec_0, t_switch, tvec_1, ys_0, ys_1):
        return np.vstack((ys_0, ys_1))

    ode_combined = OdeCombined(
        "robertson_combined", ode_1, ode_2, switch_time, switch, merge
    )

    module = Module()
    module.add(ode_1)
    module.add(ode_2)
    module.add(ode_combined)

    tvec = 0.4 * 10 ** np.arange(0, 6)
    rtol = 1e-6
    atol = np.array([1e-8, 1e-10, 1e-8])
    y0 = np.array([0, 0, 1.0])

    root_folder = tmpdir.join("root_folder").strpath
    lsoda_folder = tmpdir.join("lsoda").strpath

    w = module.compile_and_load(root_folder=root_folder, lsoda_folder=lsoda_folder)
    w.set_sec_factor(1.0)
    result_0, meta_0 = w.solve_fast_robertson_combined(y0, tvec, rtol, atol)

    assert w.get_new_traces("robertson_1") == {}
    assert w.get_new_traces("robertson_2") == {0: [[1, 1, 2]]}

    w = module.recompile_and_load(root_folder=root_folder, lsoda_folder=lsoda_folder)
    result_1, meta_1 = w.solve_fast_robertson_combined(y0, tvec, rtol, atol)
    assert w.get_new_traces("robertson_1") == {}
    assert w.get_new_traces("robertson_2") == {}

    assert np.all(result_0 == result_1)



# This file is part of sympy2c.
#
# Copyright (C) 2013-2022 ETH Zurich, Institute for Particle and Astrophysics and SIS
# ID.
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.


cdef extern from "<array>" namespace "std" nogil:
    cdef cppclass array{N} "std::array<int, {N}> ":
        array{N}() except+
        int& operator[](size_t)


cdef extern:
    void _rhs_{id_}(long * n, double * t, double * y, double * ydot)
    void _fnorm_{id_}(double * ewt, double *y)
    void _solve_fast_{id_}(double *t, double *x, double *y, int *, int)
    void _reset_next_solver_{id_}()
    void _reset_timers_{id_}()
    double _get_timer_{id_}_fast()
    double _get_timer_{id_}_lup()
    void _enable_fast_solver_{id_}(int)
    void _enable_sparse_lu_solver_{id_}(int)
    vector[vector[unsigned int]] _call_counts_{id_};


def solve_fast_lu_{id_}(double t, double[:] x, double[:] y):
    cdef int ier, i

    # copying is important as solvers will overwrite y:
    cdef np.ndarray data = np.array(y).copy()
    cdef double[:] data_view = data
    _solve_fast_{id_}(&t, &data_view[0], &data_view[0], &ier, 0)
    x[:] = data_view


def symbols_{id_}():
    return {symbols}


@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
def {id_}_dot(double lna,  double[:] y0):
    cdef np.ndarray result = np.zeros(({N},))
    cdef double [:] result_view = result
    cdef long n = {N}
    cdef double y_permuted[{N}]
    cdef double permuted_result[{N}]
    cdef int[{N}] permutation = [{permutation}]
    for j in range({N}):
        y_permuted[j] = y0[permutation[j]]
    _rhs_{id_}(&n, &lna, &y_permuted[0], &permuted_result[0])
    for j in range({N}):
        result_view[permutation[j]] = permuted_result[j]
    return result




@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
cpdef solve_fast_{id_}(
                     wrapper,
                     double[:] y0,
                     double[:] tvec,
                     rtol,
                     atol,
                     int max_iter=200000,
                     int max_order=5,
                     int enable_fast_solver=1,
                     int enable_sparse_lu_solver=1,
                     double h0=0.0,
                     ):


    assert isinstance(atol, (float, np.ndarray)), "atol must be number or numpy array"
    assert 1 <= max_order <= 5, "max_order must be in the range 1..5"

    if isinstance(atol, float):
        atol = atol * np.ones(({N},))

    if isinstance(rtol, float):
        rtol = rtol * np.ones(({N},))

    assert atol.shape == ({N},), "atol must be either a number of a vector of length {N}"
    assert rtol.shape == ({N},), "rtol must be either a number of a vector of length {N}"

    cdef int[{N}] permutation = [{permutation}]
    cdef long int neq = {N}
    cdef double __y[{N}], t, tout
    cdef int i, j
    cdef double[:] y = np.zeros(({N},))

    cdef double _atol[{N}]
    cdef double _rtol[{N}]

    for i in range({N}):
        y[i] = y0[permutation[i]]
        _atol[i] = atol[permutation[i]]
        _rtol[i] = rtol[permutation[i]]

    cdef int nt = tvec.shape[0]
    cdef double tmax = tvec[nt - 1]

    cdef double rwork[{LRW}]

    cdef long int itask = 4  # dont overshoot tcrit = rwork[0]
    rwork[0] = tmax

    cdef long int istate = 1  # must be set like this
    cdef long int iopt = 1  # optional inputs in rwork/iwork

    cdef long int itol = 4  # provide rtol and atol as vector

    cdef long int lrw = {LRW}
    cdef double lasty[{N}]

    cdef long int liw = {LIW}
    cdef long int iwork[{LIW}]
    cdef long int jt = 4  # user supplied jacobian sparse

    for i in range(liw):
        iwork[i] = 0

    for i in range(lrw):
        rwork[i] = 0.0


    # tcrit
    rwork[0] = tvec[nt - 1]
    rwork[4] = h0

    iwork[0] = iwork[1] = 0  # bandwidth sparse matrix

    iwork[5] = max_iter    # iwork(6) in fortran
    iwork[8] = max_order   # iwork(9) in fortran, max order of bdf

    cdef np.ndarray result = np.zeros((nt, {N}), order="C")
    cdef double [:, ::1] result_view = result

    cdef np.ndarray steps = np.zeros((nt,), dtype=np.uint32)
    cdef np.uint32_t [:] steps_view = steps

    result_view[0, :] = y0

    cdef long started = microtime()

    cdef int last_steps = 0

    _reset_next_solver_{id_}()
    _enable_fast_solver_{id_}(enable_fast_solver)
    _enable_sparse_lu_solver_{id_}(enable_sparse_lu_solver)
    _reset_timers_{id_}()

    _call_counts_{id_}.clear()

    cdef vector[double] hvalues_c = vector[double]()
    cdef long handle = id(list())  # random key
    _hvalues[handle] = &hvalues_c


    with nogil:
        for i in range(1, nt):
            t = tvec[i - 1]
            tout = tvec[i]

            lsodafast_(<S_fp>_rhs_{id_}, NULL, &neq, &y[0], &t, &tout, &itol, _rtol, _atol, &itask,
                &istate, &iopt, rwork, &lrw, iwork, &liw, NULL, &jt, <V_fp>_solve_fast_{id_},
                <D_fp>_fnorm_{id_}, <V_fp> _hvalue_callback, &handle)

            if istate <= 0:
                break

            for j in range({N}):
                result_view[i, permutation[j]] = y[j]

            steps_view[i] = iwork[10] - last_steps
            last_steps = iwork[10]
            iwork[10] = 0
            # step_sizes_view[i] = ls0001_.h

    cdef needed = microtime() - started

    step_sizes = []
    cdef vector[double].iterator it = hvalues_c.begin()
    while it != hvalues_c.end():
        step_sizes.append(deref(it))
        inc(it)

    _hvalues.erase(_hvalues.find(handle))

    new_traces = get_new_traces("{id_}")
    if new_traces:
        update_new_traces("{ode_id}", new_traces)

    result_meta = dict(istate=istate,
                       num_steps=iwork[10],
                       num_feval=iwork[11],
                       num_lu_solver_calls=iwork[12],
                       last_switch_at=rwork[14],
                       seconds_needed = needed/1e6,
                       steps=steps,
                       step_sizes=step_sizes,
                       new_traces=new_traces,
                       total_time_fast_solver=_get_timer_{id_}_fast(),
                       total_time_lup_solver=_get_timer_{id_}_lup(),
                       lu_call_counts=_call_counts_{id_},
                       )

    return result, result_meta



# This file is part of PyCosmo, a multipurpose cosmology calculation tool in Python.
#
# Copyright (C) 2013-2022 ETH Zurich, Institute for Particle and Astrophysics and SIS
# ID.
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.

import re

import numpy as np
import pytest

from sympy2c import (
    Function,
    Globals,
    Integral,
    InterpolationFunction1D,
    Module,
    Symbol,
    Vector,
    __version__,
)
from sympy2c.utils import align


def test_minimal(tmpdir, compile_and_load):
    x, y, z, alpha, da = map(Symbol, "x y z alpha da".split())

    v = Vector("v", 3)

    m = Module()

    m.add(Globals(y, z, alpha, v))

    _squared = InterpolationFunction1D("squared")

    m.add(Function("f", x * y + z ** _squared(alpha), x))
    m.add(Function("integral", Integral(_squared(da), da, 0, x), x))
    m.add(Function("squared", _squared(x), x))
    m.add(Function("sumv", v[0] + v[1] + v[2]))

    version = "_".join(map(str, __version__))
    assert m.get_unique_id().startswith(f"{version}_"), m.get_unique_id()

    m_loaded = compile_and_load(m)

    assert m_loaded.sympy2c_version is not None
    assert re.match(r"\d+\.\d+\.\d+", m_loaded.sympy2c_version)

    x = np.linspace(0, 1, 100)
    m_loaded.set_squared_values(x, x**2)
    assert m_loaded.squared(1) == 1
    assert abs(m_loaded.squared(0.5) - 0.5**2) < 1e-10

    m_loaded.globals.y = 1
    m_loaded.globals.z = 2
    m_loaded.globals.alpha = 0.5
    m_loaded.globals.v = np.arange(3)

    assert m_loaded.sumv() == 3.0

    assert m_loaded.globals.y == 1
    with pytest.raises(KeyError):
        m_loaded.globals.u

    print(m_loaded.globals)

    assert abs(m_loaded.f(2) - 3.189207115002721) < 1e-14
    assert abs(m_loaded.integral(1) - 1 / 3) < 1e-7

    assert m_loaded.get_unique_id() is not None


def test_from_code_snippet(tmpdir, compile_and_load):
    code = align(
        """
            |decl("a", "b", "c", "x")
            |
            |globals = [b]
            |
            |squared = InterpolationFunction1D("squared")
            |
            |f = Function("f", squared(a * b), a)
            |
            |g = Integral(f(c) * c, c, 0, b)
            |g = Function("g", g)
            |
            |h = Function("h", Integral(IfThenElse(x < 1, 1, -0), x, 0, 2))
            |i = Function("i", Min(a, c) + Max(a, c), a, c)
            """
    )

    m, _ = Module.parse_sympy_code(code)
    m_loaded = compile_and_load(m)

    x = np.linspace(0, 1, 100)
    m_loaded.set_squared_values(x, x**2)

    m_loaded.globals.b = 0.5

    assert abs(m_loaded.f(2) - 1) < 1e-8
    assert abs(m_loaded.g() - 0.5**6 / 4) < 1e-8

    assert abs(m_loaded.h() - 1) < 1e-8

    assert m_loaded.get_unique_id() is not None



# This file is part of PyCosmo, a multipurpose cosmology calculation tool in Python.
#
# Copyright (C) 2013-2022 ETH Zurich, Institute for Particle and Astrophysics and SIS
# ID.
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np

from sympy2c import Function, Module, Symbol


def test_unfunc(compile_and_load):
    a = Symbol("a")
    b = Symbol("b")

    f = Function("add", a + b, a, b)
    m = Module()
    m.add(f)
    w = compile_and_load(m)

    assert w.add(1, 2) == 3

    a_vec = np.arange(3.0)
    b_vec = np.arange(3.0)

    assert np.all(w.add(a_vec, b_vec) == a_vec + b_vec)

    a_vec = np.array((7.0,))
    b_vec = np.arange(3.0)

    assert np.all(w.add(a_vec, b_vec) == a_vec + b_vec)
    assert np.all(w.add_ufunc(a_vec, b_vec) == a_vec + b_vec)

    a_vec = np.arange(4.0)
    b_vec = np.arange(3.0)[:, None]

    assert np.all(w.add(a_vec, b_vec) == a_vec + b_vec)



# This file is part of sympy2c.
#
# Copyright (C) 2013-2022 ETH Zurich, Institute for Particle and Astrophysics and SIS
# ID.
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.


from importlib.metadata import version as _version

from .compiler import compile_if_needed_and_load
from .expressions import Max, Min, isnan
from .function import Alias, Function
from .globals import Globals
from .integral import ERROR, Checked, IfThenElse, Integral
from .interpolation import InterpolationFunction1D, InterpolationFunction1DInstance
from .module import Module
from .ode import Ode
from .ode_combined import OdeCombined
from .ode_fast import OdeFast
from .python_function import PythonFunction
from .symbol import Symbol, symbols
from .vector import Vector, VectorElement

__all__ = [
    "compile_if_needed_and_load",
    "Max",
    "Min",
    "isnan",
    "Alias",
    "Function",
    "Globals",
    "ERROR",
    "Checked",
    "IfThenElse",
    "Integral",
    "InterpolationFunction1D",
    "InterpolationFunction1DInstance",
    "Module",
    "Ode",
    "OdeCombined",
    "OdeFast",
    "PythonFunction",
    "Symbol",
    "symbols",
    "Vector",
    "VectorElement",
]


__author__ = "Uwe Schmitt"
__email__ = "uwe.schmitt@id.ethz.ch"
__version__ = _version(__package__)



# This file is part of sympy2c.
#
# Copyright (C) 2013-2022 ETH Zurich, Institute for Particle and Astrophysics and SIS
# ID.
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.

import sympy as sp

from .symbol import Symbol
from .utils import align


class VectorElement(sp.Symbol):
    def __new__(cls, name, index=None):
        # index is not defined in base classes __new__ and as such  not provided when
        # unpickling, so we need default value to make unpickling work. when we unpickle
        # name is already of formm "v[i]".
        if index is not None:
            return super().__new__(cls, "{}[{}]".format(name, index))
        else:
            return super().__new__(cls, name)

    def c_args_decl(self):
        return self

    def __str__(self):
        return self.name

    def _shrinked(self):
        # name + index without the brackets. a[0] -> a0
        return self.name.replace("[", "").replace("]", "")

    @property
    def index(self):
        ix_part = self.name.split("[")[1][:-1]
        return int(ix_part)

    @property
    def vector_name(self):
        return self.name.split("[")[0]

    def as_function_arg(self):
        return Symbol(self._shrinked())

    def cython_decl(self):
        return "double " + self._shrinked()

    def c_decl(self):
        return "double *" + self._shrinked()

    def __getstate__(self):
        return sp.Symbol.__getstate__(self), self.parent

    def __setstate__(self, data):
        state, self.parent = data

    @property
    def free_symbols(self):
        return set((self,))

    def __hash__(self):
        return hash(str(self))

    def __eq__(self, other):
        if type(self) != type(other):  # noqa E271
            return False
        return self.name == other.name


class Vector(sp.Symbol):
    def __new__(cls, name, size=None):
        # size is not defined in base classes __new__ and as such  not provided when
        # unpickling, so we need default value to make unpickling work. when we unpickle
        # name is already of formm "v[i]".
        assert not name.startswith(
            "_"
        ), "you must not use underscore as first character"

        if size is not None:
            return super().__new__(cls, "{}[{}]".format(name, size))
        else:
            return super().__new__(cls, name)

    is_vector = True

    def __str__(self):
        return self.name

    def _name(self):
        return self.name.split("[")[0]

    def _size(self):
        number = self.name.split("[")[1].lstrip("[").rstrip("]")
        return int(number)

    __len__ = _size

    def __getitem__(self, i):
        if not 0 <= i < self._size():
            raise IndexError()
        result = VectorElement(self._name(), i)
        result.parent = self
        return result

    def as_function_arg(self):
        return self._name()
        return "&{}[0]".format(self._name())

    def cython_decl(self):
        return "double [::1] {}".format(self._name())

    def c_decl(self):
        return f"double * {self._name()}"

    def ctypes_decl(self):
        # numba does not support:
        # return "np.ctypeslib.ndpointer(dtype=np.double, ndim=1, flags='CONTIGUOUS')"
        return "ctypes.POINTER(ctypes.c_double)"

    def numba_type_decl(self):
        return "float64[:]"

    @property
    def free_symbols(self):
        elements = (VectorElement(self._name(), i) for i in range(self._size()))
        return set(elements)

    def __hash__(self):
        return hash(str(self))

    def __eq__(self, other):
        if type(self) != type(other):  # noqa E271
            return False

        return self.name == other.name

    def setter_c_code(self):
        return align(
            """
            |extern "C" void _set_{_name}(double *_v) {{
            |    memcpy({_name}, _v, {n} * sizeof(double));
            |}}
            """
        ).format(_name=self._name(), n=self._size())

    def getter_c_code(self):
        return align(
            """
            |extern "C" void _get_{_name}(double *result) {{
            |    memcpy(result, {_name}, {n} * sizeof(double));
            |}}
            """
        ).format(_name=self._name(), n=self._size())

    def setter_header_code(self):
        return "void _set_{_name}(double *); ".format(_name=self._name())

    def getter_header_code(self):
        return "void _get_{_name}(double *); ".format(_name=self._name())

    setter_cython_header_code = setter_header_code
    getter_cython_header_code = getter_header_code

    def cython_globals_setter_code(self):
        return (
            align(
                """
           |if "{_name}" in _g:
           |    values_1d = np.array(_g["{_name}"], dtype=np.double, order="c")
           |    if values_1d.shape[0] != {n}:
           |         raise ValueError("value must be list of 1d array of length {n}")
           |    _set_{_name}(&values_1d[0])
           """
            )
            .format(_name=self._name(), n=self._size())
            .rstrip()
        )

    def cython_globals_getter_code(self):
        return align(
            """
           |values_1d = np.zeros(({n},), dtype=np.double)
           |_get_{_name}(&values_1d[0])
           |result["{_name}"] = np.array(values_1d)
           """
        ).format(_name=self._name(), n=self._size())



# This file is part of sympy2c.
#
# Copyright (C) 2013-2022 ETH Zurich, Institute for Particle and Astrophysics and SIS
# ID.
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.

def symbols_{id_}():
    return {symbols}



def solve_fast_{id_}(wrapper,
                     np.ndarray y0,
                     np.ndarray tvec,
                     rtol,
                     atol,
                     int max_iter=200000,
                     int max_order=5,
                     enable_fast_solver=1,
                     enable_sparse_lu_solver=1,
                     double h0=0.0,
                     ):

{switch_time_function_src}

{switch_function_src}

{merge_function_src}

    assert y0.ndim == 1, "y0 must be one-dimensional"
    assert tvec.ndim == 1, "tvec must be one-dimensional"

    cdef double t_switch = {switch_time_function_name}(wrapper, tvec)

    if tvec[0] < t_switch < tvec[-1]:
        tvec_0 = np.concatenate((tvec[tvec < t_switch], [t_switch]))
        tvec_1 = np.concatenate(([t_switch], tvec[tvec > t_switch]))
    elif tvec[-1] <= t_switch:
        tvec_0 = tvec
        tvec_1 = np.arange(0, dtype=tvec.dtype)
    else:
        tvec_0 = np.arange(0, dtype=tvec.dtype)
        tvec_1 = tvec

    n = y0.shape[0]

    if t_switch > tvec[0]:
        result_0, meta_0 = {solver_0}(wrapper, y0, tvec_0, rtol, atol, max_iter,
                                      max_order, enable_fast_solver,
                                      enable_sparse_lu_solver, h0)
        y0 = {switch_function_name}(wrapper, t_switch, result_0)
    else:
        result_0 = np.zeros((0, len(y0)), dtype=y0.dtype)
        meta_0 = dict(istate=2, num_steps=0, num_feval=0,
                    num_lu_solver_calls=0, last_switch_at=t_switch, seconds_needed=0.0,
                    total_time_fast_solver=0.0, total_time_lup_solver=0.0,
                    steps=[], step_sizes=[], new_traces=[], lu_call_counts=0,
                    )
        y0 = {switch_function_name}(wrapper, t_switch, y0.reshape(1, -1))

    # we always call solver_1 to make sure result_1 and meta_1 are populated,
    # might be that this solver does no single step
    if t_switch < tvec[-1]:
        result_1, meta_1 = {solver_1}(wrapper, y0, tvec_1, rtol, atol, max_iter,
                                      max_order, enable_fast_solver,
                                      enable_sparse_lu_solver, h0)

    else:
        result_1 = np.zeros((0, len(y0)), dtype=y0.dtype)
        meta_1 = dict(istate=2, num_steps=0, num_feval=0,
                     num_lu_solver_calls=0, last_switch_at=tvec[0], seconds_needed=0.0,
                     total_time_fast_solver=0.0, total_time_lup_solver=0.0,
                     steps=[], step_sizes=[], new_traces=[], lu_call_counts=0,
                     )

    meta = dict(istate=min(meta_0["istate"], meta_1["istate"]),
                num_steps=meta_0["num_steps"] + meta_1["num_steps"],
                num_feval=meta_0["num_feval"] + meta_1["num_feval"],
                num_lu_solver_calls=meta_0["num_lu_solver_calls"] + meta_1["num_lu_solver_calls"],
                last_switch_at=max(meta_0["last_switch_at"], meta_1["last_switch_at"]),
                seconds_needed=meta_0["seconds_needed"] + meta_1["seconds_needed"],
                total_time_fast_solver=meta_0["total_time_fast_solver"] + meta_1["total_time_fast_solver"],
                total_time_lup_solver=meta_0["total_time_lup_solver"] + meta_1["total_time_lup_solver"],
                steps=np.concatenate((meta_0["steps"], meta_1["steps"])),
                step_sizes=np.concatenate((meta_0["step_sizes"], meta_1["step_sizes"])),
                new_traces=[meta_0["new_traces"], meta_1["new_traces"]],
                lu_call_counts=[meta_0["lu_call_counts"], meta_1["lu_call_counts"]],
                )

    if t_switch < tvec[-1]:
        result_0 = result_0[:-1]
        tvec_0 = tvec_0[:-1]
    if t_switch not in tvec and t_switch > tvec[0]:
        result_1 = result_1[1:]
        tvec_1 = tvec_1[1:]

    return {merge_function_name}(wrapper, tvec_0, t_switch, tvec_1, result_0, result_1), meta



#!/usr/bin/env python
from numba import njit

from sympy2c import Function, Module, symbols


def test_numba(compile_and_load):
    a, b, c, d, e = symbols("a b c d e")

    m = Module()
    f = Function("f", a + b, a, b)
    g = Function("g", a**2, a)
    m.add(f)
    m.add(g)

    w = compile_and_load(m)
    f = w.f
    g = w.g

    assert g(1.0) == 1.0
    assert f(1.0, 2.0) == 3.0

    @njit
    def compute():
        return f(1.0, 2.0) + g(3.0)

    assert compute() == 12.0



# This file is part of sympy2c.
#
# Copyright (C) 2013-2022 ETH Zurich, Institute for Particle and Astrophysics and SIS
# ID.
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.

import inspect
import pathlib
from hashlib import md5
from textwrap import dedent, indent

from .ode_fast import OdeFast, concat_generator_results

CYTHON_TEMPLATE = "ode_combined_cython_template.pyx"


class OdeCombined:
    def __init__(self, name, ode_0, ode_1, switch_time, switch, merge):
        assert isinstance(ode_0, OdeFast), "only works for fast ode"
        assert isinstance(ode_1, OdeFast), "only works for fast ode"
        self.name = name
        self.odes = [ode_0, ode_1]
        self._unique_id = None

        self.switch_time_function_src = indent(
            dedent(inspect.getsource(switch_time)), "    "
        )
        self.switch_function_src = indent(dedent(inspect.getsource(switch)), "    ")
        self.merge_function_src = indent(dedent(inspect.getsource(merge)), "    ")

        self.switch_time_function_name = switch_time.__name__
        self.switch_function_name = switch.__name__
        self.merge_function_name = merge.__name__

    def __str__(self):
        return "<OdeCombined {}, {}, {}, {}, {}>".format(
            self.name, self.odes, self.switch_time, self.switch, self.merge
        )

    @property
    def wrapper(self):
        return OdeCombinedWrapper

    def get_unique_id(self):
        if self._unique_id is None:
            hasher = md5()
            hasher.update(self.name.encode("utf-8"))
            for ode in self.odes:
                hasher.update(ode.get_unique_id().encode("utf-8"))
            hasher.update(self.switch_time_function_src.encode("utf-8"))
            hasher.update(self.switch_function_src.encode("utf-8"))
            hasher.update(self.merge_function_src.encode("utf-8"))
            hasher.update(self.switch_time_function_name.encode("utf-8"))
            hasher.update(self.switch_function_name.encode("utf-8"))
            hasher.update(self.merge_function_name.encode("utf-8"))
            self._unique_id = hasher.hexdigest()
        return self._unique_id


class OdeCombinedWrapper:
    def __init__(self, ode_combined, globals_, visitor):
        self.ode_combined = ode_combined
        self.odes = ode_combined.odes
        self.globals_ = globals_
        self.visitor = visitor
        self.name = ode_combined.name
        self._unique_id = None

    def setup_code_generation(self):
        pass

    def get_unique_id(self):
        if self._unique_id is None:
            hasher = md5()
            hasher.update(self.ode_combined.get_unique_id().encode("utf-8"))
            # for level, level_traces in sorted(self.traces.items()):
            #    hasher.update(str(sorted(level_traces)).encode("utf-8"))
            self._unique_id = hasher.hexdigest()
        return self._unique_id

    def determine_required_extra_wrappers(self):
        return
        for ode in self.odes:
            for rhs in ode.rhs:
                self.visitor.visit(rhs)

    @concat_generator_results
    def c_header(self):
        yield f"PT get_new_traces_{self.name}();"

    @concat_generator_results
    def c_code(self, header_file_path):
        yield f"PT _new_traces_{self.name};"

    @concat_generator_results
    def cython_code(self, header_file_path):
        here = pathlib.Path(__file__).parent
        template = (here / CYTHON_TEMPLATE).read_text()

        symbols = repr(list(map(repr, self.odes[0].lhs)))

        yield template.format(
            id_=self.name,
            symbols=symbols,
            solver_0="solve_fast_{}".format(self.odes[0].name),
            solver_1="solve_fast_{}".format(self.odes[1].name),
            switch_time_function_src=self.ode_combined.switch_time_function_src,
            switch_function_src=self.ode_combined.switch_function_src,
            merge_function_src=self.ode_combined.merge_function_src,
            switch_time_function_name=self.ode_combined.switch_time_function_name,
            switch_function_name=self.ode_combined.switch_function_name,
            merge_function_name=self.ode_combined.merge_function_name,
        )



# This file is part of PyCosmo, a multipurpose cosmology calculation tool in Python.
#
# Copyright (C) 2013-2022 ETH Zurich, Institute for Particle and Astrophysics and SIS
# ID.
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.


import pytest
import sympy

from sympy2c.rec_inv import rec_inv


@pytest.mark.parametrize(
    "n, m",
    [
        (2, 0),
        (2, 1),
        (2, 2),
        (2, 3),
        (2, 4),
        (3, 0),
        (3, 1),
        (3, 2),
        (3, 3),
        (3, 4),
        (3, 5),
        (4, 2),
    ],
)
def test_recinv(m, n):
    a, b, c = sympy.symbols("a b c")
    A = sympy.eye(n)
    for i in range(n - 1):
        A[i, i + 1] = a if i % 2 == 0 else a * b
        A[i + 1, i] = b if i % 2 == 0 else a * b + c
        # A[0, i] = b + c + 1

    B = sympy.zeros(m, n)
    D = sympy.eye(m)

    M = sympy.Matrix(sympy.BlockMatrix([[A, B.T], [B, D]]))

    xii, Mi = rec_inv(M)
    check = M * Mi
    while any(str(a).startswith("__x") for a in check.atoms()):
        check = check.subs(xii)
    assert sympy.simplify(check) == sympy.eye(m + n)



# This file is part of sympy2c.
#
# Copyright (C) 2013-2022 ETH Zurich, Institute for Particle and Astrophysics and SIS
# ID.
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.

import multiprocessing

from sympy import Matrix


def _jacobian(args):
    expressions, variables = args
    return expressions.jacobian(variables)


def jacobian(expressions, variables, N=4):
    if N == 1 or expressions.rows < 4 * N:
        return expressions.jacobian(variables)

    blocks = [Matrix(expressions[i::N]) for i in range(N)]
    args = [(b, variables) for b in blocks]
    with multiprocessing.Pool(N) as pool:
        partial_results = pool.map(_jacobian, args)

    result = [partial_results[i % N][i // N, :] for i in range(expressions.rows)]
    return Matrix(result)



# This file is part of sympy2c.
#
# Copyright (C) 2013-2022 ETH Zurich, Institute for Particle and Astrophysics and SIS
# ID.
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.

from hashlib import md5

import sympy as sp
from sympy.matrices.matrices import MatrixBase

from .create_curry_class import create_curry_class
from .utils import align, concat_generator_results
from .wrapper import WrapperBase


def Integral(integrand, integration_variable, low, high, id_="default"):
    if "[" in str(integration_variable):
        raise ValueError("dont use array or array elements as integration variable")
    if not isinstance(low, (int, float)):
        assert integration_variable not in low.free_symbols

    if not isinstance(high, (int, float)):
        assert integration_variable not in high.free_symbols

    if isinstance(integrand, MatrixBase):
        integrands = list(integrand)
        assert len(integrands) == 1, "no implementation for vector valued integrals"
        integrand = integrands[0]

    if isinstance(low, (int, float)):
        low = sp.Number(low)
    if isinstance(high, (int, float)):
        high = sp.Number(high)

    integrand = Integrand(integrand, integration_variable)

    unique_id = md5(
        (
            str(integrand.sort_key() + low.sort_key() + high.sort_key())
            + str(integration_variable)
            + str(id_)
        ).encode("utf-8")
    ).hexdigest()

    # all args must be "sympy"-fiable:
    result = _Integral(
        integrand, integration_variable, low, high, id_, sp.Symbol(unique_id)
    )
    return result


class _Integral(sp.Function("Integral", nargs=6, real=True)):
    @property
    def free_symbols(self):
        integrand, integration_variable, low, high, id_ = self.args[:-1]
        return integrand.free_symbols | low.free_symbols | high.free_symbols

    def get_unique_id(self):
        return self.args[-1]

    def atoms(self, what=None):
        return (
            self.integrand.atoms(what)
            | self.low.atoms(what)
            | self.high.atoms(what)
            | self.integration_variable.atoms(what)
        )

    @property
    def integrand(self):
        return self.args[0]

    @property
    def integration_variable(self):
        return self.args[1]

    @property
    def low(self):
        return self.args[2]

    @property
    def high(self):
        return self.args[3]


class Integrand(sp.Expr):
    def __init__(self, expression, integration_variable):
        self.expression = expression
        self.integration_variable = integration_variable

    @property
    def free_symbols(self):
        return self.expression.free_symbols - {self.integration_variable}

    def atoms(self, what=None):
        return self.expression.atoms(what) | self.integration_variable.atoms(what)

    def __getstate__(self):
        return (self.expression, self.integration_variable)

    def __setstate__(self, data):
        self.expression, self.integration_variable = data


IfThenElse = sp.Function("IfThenElse", nargs=3, real=True)


class ERROR(sp.Expr):
    def __new__(clz, message=None):
        # default value None required for unpickling!
        # we can not implement __init__ because sympy does some magick which
        # avoids setting new attributes within __init__
        instance = super().__new__(clz)
        instance.message = message
        return instance

    def __str__(self):
        return 'ERROR("{}")'.format(self.message)

    def __getstate__(self):
        return self.__dict__

    def __setstate__(self, dd):
        self.__dict__.update(dd)


def Checked(expr, message):
    return IfThenElse(expr, expr.lhs, ERROR(message))


class IntegralFunctionWrapper(WrapperBase):
    seen = set()

    @classmethod
    def reset(clz):
        clz.seen = set()

    def __init__(self, name, free_vars, integrand_name, id_):
        self.name = name
        self.free_vars = free_vars
        # self.low = low
        # self.high = high
        self.integrand_name = integrand_name
        self.id_ = id_

    def setup_code_generation(self):
        pass

    def c_header(self):
        args = ", ".join(
            "double {}".format(var) for var in self.free_vars + ["_low", "_high"]
        )

        return "double {name}({args});".format(name=self.name, args=args)

    def cython_code(self, header_file_path):
        return ""

    @concat_generator_results
    def c_code(self, header_file_path):
        curry_class_name, curry_class_code = create_curry_class(len(self.free_vars))

        # don't craeate same currying abstract base class multiple
        # times:
        if curry_class_name not in self.seen:
            self.seen.add(curry_class_name)
            yield curry_class_code

        args = ", ".join(
            "double {}".format(var) for var in self.free_vars + ["_low", "_high"]
        )

        # we insert "_" at beginning of c functions to avoid clash with
        # cython functions:
        cons_args = ", ".join(
            ["_" + self.integrand_name] + list(map(str, self.free_vars))
        )

        yield align(
            """
        |inline double {name}({args}) {{
        |    {curry_class_name} curried_function({cons_args});
        |    return quadpack_integration(&curried_function, "{id_}", _low, _high);
        |}}
        |"""
        ).format(
            name=self.name,
            id_=self.id_,
            # low=self.low,
            # high=self.high,
            args=args,
            curry_class_name=curry_class_name,
            cons_args=cons_args,
        )

    def determine_required_extra_wrappers(self):
        pass



# This file is part of PyCosmo, a multipurpose cosmology calculation tool in Python.
#
# Copyright (C) 2013-2022 ETH Zurich, Institute for Particle and Astrophysics and SIS
# ID.
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.


import numpy as np
import pytest
import scipy.special
import sympy

from sympy2c import Function, Module, symbols
from sympy2c.expressions import hyper_2f1


@pytest.mark.parametrize("name", "ijky")
def test_bessel(name, tmpdir, compile_and_load):
    x, n = symbols("x n")

    sympy_fun = getattr(sympy, "bessel" + name)
    scipy_fun = getattr(scipy.special, name + "v")

    fun = Function("fun", sympy_fun(n * n, x * x), n, x)
    m = Module()
    m.add(fun)
    w = compile_and_load(m)

    assert np.allclose(w.fun(0, 1), scipy_fun(0, 1))
    assert np.allclose(w.fun(0, 2), scipy_fun(0, 4))
    assert np.allclose(w.fun(1, 2), scipy_fun(1, 4))


def test_hyper(tmpdir, compile_and_load):
    a, b, c, x = symbols("a b c x")

    h = Function("h", hyper_2f1(a, b, c, x), a, b, c, x)

    m = Module()
    m.add(h)

    w = compile_and_load(m)

    for args in [(0.1, 1, 1, 0), (2, 1, 4, 1)]:
        assert np.allclose(w.h(*args), scipy.special.hyp2f1(*args))


def test_gamma(tmpdir, compile_and_load):
    (x,) = symbols("x")

    h = Function("h", sympy.gamma(x), x)

    m = Module()
    m.add(h)

    w = compile_and_load(m)

    for x in (1, 2, 3, 3.5):
        assert np.allclose(w.h(x), scipy.special.gamma(x))



# This file is part of PyCosmo, a multipurpose cosmology calculation tool in Python.
#
# Copyright (C) 2013-2022 ETH Zurich, Institute for Particle and Astrophysics and SIS
# ID.
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.


import numpy as np
import pytest

from sympy2c import Globals, Module, Ode, Symbol, Vector
from sympy2c.utils import align


def test_ode_solver(tmpdir, compile_and_load):
    v1x, v2x, v1y, v2y = map(Symbol, "v1x v2x v1y v2y".split())
    p1x, p2x, p1y, p2y = map(Symbol, "p1x p2x p1y p2y".split())

    m1, m2 = map(Symbol, "m1 m2".split())

    t = Symbol("t")

    p1x_dot = v1x
    p1y_dot = v1y

    p2x_dot = v2x
    p2y_dot = v2y

    d2 = ((p1x - p2x) ** 2 + (p1y - p2y) ** 2) ** 0.5

    v1x_dot = -m2 * (p1x - p2x) / d2**3
    v1y_dot = -m2 * (p1y - p2y) / d2**3

    v2x_dot = -m1 * (p2x - p1x) / d2**3
    v2y_dot = -m1 * (p2y - p1y) / d2**3

    ode = Ode(
        "planets",
        t,
        [p1x, p1y, p2x, p2y, v1x, v1y, v2x, v2y],
        [p1x_dot, p1y_dot, p2x_dot, p2y_dot, v1x_dot, v1y_dot, v2x_dot, v2y_dot],
        compute_jacobian=True,
    )

    module = Module()
    module.add(Globals(m1, m2))
    module.add(ode)

    _check(module, tmpdir, True, compile_and_load)
    _check(module, tmpdir, False, compile_and_load)

    assert module.get_unique_id() is not None


def _check(module, tmpdir, use_jacobian, compile_and_load):
    wrapper = compile_and_load(module)

    assert wrapper.get_unique_id() is not None

    wrapper.globals.m1 = 2
    wrapper.globals.m2 = 1

    p1 = (5, 0)
    v1 = (0.2, -0.2)

    p2 = (10, 0)
    v2 = (0, 0.5)

    t = np.linspace(0, 60, 1000)

    assert wrapper.solve_planets.__doc__.startswith(
        "solve_planets(y0, tgrid, atol, rtol, max_iter, max_order, use_jacobian)"
    )

    y, meta = wrapper.solve_planets(
        np.array((p1[0], p1[1], p2[0], p2[1], v1[0], v1[1], v2[0], v2[1]), dtype=float),
        t,
        1e-5,
        1e-5,
        use_jacobian=use_jacobian,
    )

    assert isinstance(meta, dict)

    # unpack columns:
    p1x, p1y, p2x, p2y, v1x, v1y, v2x, v2y = y.T

    is_ = [
        p1x.mean(),
        p1y.mean(),
        p2x.mean(),
        p2y.mean(),
        p1x.std(),
        p1y.std(),
        p2x.std(),
        p2y.std(),
    ]

    tobe = [
        10.419238674733615,
        1.4668800402967992,
        11.161522650532781,
        0.06623991940640099,
        2.6002221952237,
        1.3737285646379265,
        2.885845577081757,
        2.0178019356054047,
    ]

    rel_diffs = [abs(i - t) / abs(t) for (i, t) in zip(is_, tobe)]
    assert all([rel_diff < 1e-8 for rel_diff in rel_diffs])

    assert wrapper.symbols_planets() == [
        "p1x",
        "p1y",
        "p2x",
        "p2y",
        "v1x",
        "v1y",
        "v2x",
        "v2y",
    ]


def test_ode_declaration(tmpdir, compile_and_load):
    code = align(
        """
    |decl("p1x", "p2x", "p1y", "p2y",
    |     "v1x", "v2x", "v1y", "v2y",
    |     "t")
    |
    |m1, m2 = map(Symbol, "m1 m2".split())
    |
    |globals = [m1, m2]
    |
    |p1x_dot = v1x
    |p1y_dot = v1y
    |
    |p2x_dot = v2x
    |p2y_dot = v2y
    |
    |d2 = ((p1x - p2x) ** 2 + (p1y - p2y) ** 2) ** .5
    |
    |v1x_dot = -m2 * (p1x - p2x) / d2 ** 3
    |v1y_dot = -m2 * (p1y - p2y) / d2 ** 3
    |
    |v2x_dot = -m1 * (p2x - p1x) / d2 ** 3
    |v2y_dot = -m1 * (p2y - p1y) / d2 ** 3
    |
    |Ode(
    |    "planets",
    |    t,
    |    [p1x, p1y, p2x, p2y, v1x, v1y, v2x, v2y],
    |    [p1x_dot, p1y_dot, p2x_dot, p2y_dot, v1x_dot, v1y_dot, v2x_dot, v2y_dot],
    |    compute_jacobian=False
    |)
    |
    |"""
    )

    module, _ = Module.parse_sympy_code(code)
    assert module.get_unique_id() is not None
    _check(module, tmpdir, False, compile_and_load)
    with pytest.raises(AssertionError) as e:
        _check(module, tmpdir, True, compile_and_load)

    assert "compute_jacobian" in str(e.value), str(e.value)


def test_rorbertsons_reaction_example(tmpdir, compile_and_load):
    # school book example for numerical issues with a stiff ODE.

    y = Vector("yi", 3)
    t = Symbol("t")

    ydot = [None, None, None]
    ydot[0] = 1.0e4 * y[1] * y[2] - 0.04e0 * y[0]
    ydot[2] = 3.0e7 * y[1] * y[1]
    ydot[1] = -1.0 * (ydot[0] + ydot[2])

    ode = Ode("robertson", t, y, ydot, compute_jacobian=False)
    module = Module()
    module.add(ode)
    assert module.get_unique_id() is not None
    wrapper = compile_and_load(module)
    assert wrapper.get_unique_id() is not None

    y0 = np.array((1.0, 0.0, 0.0))

    tvec = 0.4 * 10 ** np.arange(1.0, 13.0)

    y_numerical, _ = wrapper.solve_robertson(y0, tvec, 1e-5, 1e-5, use_jacobian=False)

    tobe = np.array(
        [
            [1.00000000e00, 0.00000000e00, 0.00000000e00],
            [7.26396624e-01, 9.60786191e-06, 2.73593768e-01],
            [4.51749939e-01, 3.23854684e-06, 5.48246822e-01],
            [1.83279161e-01, 8.94691224e-07, 8.16719945e-01],
            [3.89748728e-02, 1.62139825e-07, 9.61024965e-01],
            [4.93229552e-03, 1.98257581e-08, 9.95067685e-01],
            [5.13206276e-04, 2.05385553e-09, 9.99486792e-01],
            [5.25780257e-05, 2.10322585e-10, 9.99947422e-01],
            [4.79111031e-06, 1.91645284e-11, 9.99995209e-01],
            [-1.99506786e-07, -7.98020732e-13, 1.00000020e00],
            [1.12615228e-06, 4.50460967e-12, 9.99998874e-01],
            [-5.71582448e-06, -2.28633004e-11, 1.00000572e00],
        ]
    )

    assert np.allclose(y_numerical, tobe)

    ode = Ode("robertson", t, y, ydot, compute_jacobian=True)
    module = Module()
    module.add(ode)
    assert module.get_unique_id() is not None

    wrapper = compile_and_load(module)
    assert wrapper.get_unique_id() is not None

    rtol = 1e-4
    atol = np.array([1e-6, 1e-10, 1e-6])

    y, _ = wrapper.solve_robertson(y0, tvec, rtol, atol, use_jacobian=True)

    tobe = np.array(
        [
            [1.00000000e00, 0.00000000e00, 0.00000000e00],
            [7.26427754e-01, 9.60916477e-06, 2.73562637e-01],
            [4.51778459e-01, 3.23889933e-06, 5.48218302e-01],
            [1.83293589e-01, 8.94748423e-07, 8.16705516e-01],
            [3.89904526e-02, 1.62207632e-07, 9.61009385e-01],
            [4.93640957e-03, 1.98424000e-08, 9.95063571e-01],
            [5.16183750e-04, 2.06578837e-09, 9.99483814e-01],
            [5.17980795e-05, 2.07202864e-10, 9.99948202e-01],
            [5.28360449e-06, 2.11345258e-11, 9.99994716e-01],
            [4.65880673e-07, 1.86352351e-12, 9.99999534e-01],
            [1.42835178e-08, 5.71341234e-14, 9.99999986e-01],
            [1.28379931e-07, 5.13519724e-13, 9.99999872e-01],
        ]
    )

    assert np.allclose(y, tobe)



# This file is part of sympy2c.
#
# Copyright (C) 2013-2022 ETH Zurich, Institute for Particle and Astrophysics and SIS
# ID.
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.

import json
import os
import pathlib
from hashlib import md5

import portalocker
from sympy import Matrix

from .function import expression_hash
from .lu_generator import argsort, generate_code, setup_code_generation
from .utils import (
    concat_generator_results,
    create_folder_if_not_exists,
    sympy2c_cache_folder,
)
from .wrapper import WrapperBase

CYTHON_TEMPLATE = "ode_fast_cython_template.pyx"


def known_traces_file(unique_id):
    folder = os.path.join(sympy2c_cache_folder(), "known_traces")
    create_folder_if_not_exists(folder)
    return os.path.join(folder, unique_id + ".json")


def _load_traces(path):
    with portalocker.Lock(path, "r", timeout=1, encoding="utf-8") as fh:
        return {int(k): v for (k, v) in json.load(fh).items()}


def _write_json(path, data):
    with portalocker.Lock(path, "w", encoding="utf-8", flags=portalocker.LOCK_EX) as fh:
        json.dump(data, fh, indent=4)
        fh.flush()


def read_known_traces(unique_id):
    path = known_traces_file(unique_id)
    if not os.path.exists(path):
        return {}
    return _load_traces(path)


def write_known_traces(unique_id, traces):
    path = known_traces_file(unique_id)
    with portalocker.Lock(path, "w", encoding="utf-8", flags=portalocker.LOCK_EX) as fh:
        json.dump(traces, fh, indent=4)
        fh.flush()


def update_new_traces(unique_id, new_traces):
    path = new_traces_file(unique_id)
    with portalocker.Lock(
        path + ".lock", "w", encoding="utf-8", flags=portalocker.LOCK_EX
    ) as fh:
        if os.path.exists(path):
            with open(path, "r") as fh:
                traces = json.load(fh)
        else:
            traces = {}

        for k, v in new_traces.items():
            traces.setdefault(k, []).extend(
                vi for vi in v if vi not in traces.get(k, [])
            )
        with open(path, "w") as fh:
            json.dump(traces, fh, indent=4)


def new_traces_file(unique_id):
    folder = os.path.join(
        os.environ.get("FAKE_CACHE_FOLDER", sympy2c_cache_folder()), "new_traces"
    )
    create_folder_if_not_exists(folder)

    return os.path.join(folder, unique_id + ".json")


def read_new_traces(unique_id):
    path = new_traces_file(unique_id)
    if not os.path.exists(path):
        return {}
    traces = _load_traces(path)
    try:
        os.unlink(path)
    except IOError:
        if os.path.exists(path):
            raise

    return traces


class OdeFast:
    def __init__(self, name, t, lhs, rhs, splits=None, reorder=False):
        assert len(lhs) == len(rhs)
        self.name = name
        self.t = t
        self.lhs = lhs
        self.rhs = Matrix(rhs)
        if not splits:
            splits = [len(lhs)]

        if max(splits) > len(rhs):
            raise ValueError(
                "split {} is beyond size of the system {}".format(max(splits), len(rhs))
            )
        self.splits = splits
        self.reorder = reorder
        self._unique_id = None
        self._unique_id_equations = None

    def __str__(self):
        return "<OdeFast {} {} {} {}, {}, {}>".format(
            self.name, self.t, self.lhs, self.rhs, self.splits, self.reorder
        )

    @property
    def wrapper(self):
        return OdeFastWrapper

    def get_unique_id(self):
        if self._unique_id is None:
            hasher = md5()
            hasher.update(self.name.encode("utf-8"))
            hasher.update(self.get_unique_id_equations().encode("utf-8"))
            hasher.update(str(self.splits).encode("utf-8"))
            self._unique_id = hasher.hexdigest()
        return self._unique_id

    def get_unique_id_equations(self):
        if self._unique_id_equations is None:
            hasher = md5()
            hasher.update(str(self.t).encode("utf-8"))
            hasher.update(str(self.lhs).encode("utf-8"))
            hasher.update(str(self.reorder).encode("utf-8"))
            for ri in self.rhs:
                hasher.update(expression_hash(ri))
            self._unique_id_equations = hasher.hexdigest()
        return self._unique_id_equations


class OdeFastWrapper(WrapperBase):
    def __init__(self, ode, globals_, visitor, traces=None):
        self.ode = ode
        self.globals_ = globals_
        self.visitor = visitor
        self.t = ode.t
        self.rhs = ode.rhs
        self.lhs = ode.lhs
        self.name = ode.name
        self.reorder = ode.reorder
        if traces is None:
            traces = read_known_traces(ode.get_unique_id())
        self.traces = traces
        self.splits = ode.splits
        self.n = len(ode.lhs)
        self._unique_id = None
        self._c_code = None

    def __str__(self):
        hash_ = md5(str(self.traces).encode("utf-8")).hexdigest()
        return "<OdeFastWrapper {} {} {} {} {}>".format(
            self.name, self.t, self.lhs, self.rhs, hash_
        )

    def compute_permutation(self):
        if self.reorder:
            state_vars = set(self.lhs)
            counts = [-len(rhsi.atoms() & state_vars) for rhsi in self.rhs]
            permutation = argsort(counts)
        else:
            permutation = list(range(self.n))
        return tuple(permutation)

    def setup_code_generation(self):
        self.permutation = self.compute_permutation()

        rhs = Matrix([self.rhs[p] for p in self.permutation])
        lhs = [self.lhs[p] for p in self.permutation]

        cache, J, h, el0, M, rhs = setup_code_generation(
            rhs,
            self.t,
            lhs,
            self.visitor,
            self.ode.get_unique_id_equations(),
            self.permutation,
        )
        self._c_code = list(
            generate_code(
                self.name,
                rhs,
                lhs,
                self.visitor,
                self.traces,
                self.splits,
                cache,
                J,
                h,
                el0,
                M,
                self.permutation,
            )
        )

    def get_unique_id(self):
        if self._unique_id is None:
            hasher = md5()
            hasher.update(self.ode.get_unique_id().encode("utf-8"))
            for level, level_traces in sorted(self.traces.items()):
                hasher.update(str(sorted(level_traces)).encode("utf-8"))
            self._unique_id = hasher.hexdigest()
        return self._unique_id

    def determine_required_extra_wrappers(self):
        for rhs in self.rhs:
            self.visitor.visit(rhs)

    def c_header(self):
        return ""

    @concat_generator_results
    def c_code(self, header_file_path):
        yield from self._c_code

    @concat_generator_results
    def cython_code(self, header_file_path):
        here = pathlib.Path(__file__).parent
        template = (here / CYTHON_TEMPLATE).read_text()

        liw = 20 + self.n

        lrw = 22 + 16 * self.n + self.n  # extra for lasty storate

        symbols = repr(list(map(repr, self.lhs)))
        permutation = ", ".join(map(str, self.permutation))

        yield template.format(
            N=self.n,
            id_=self.name,
            LRW=lrw,
            LIW=liw,
            symbols=symbols,
            ode_id=self.ode.get_unique_id(),
            permutation=permutation,
            n_splits=len(self.splits),
        )



# This file is part of sympy2c.
#
# Copyright (C) 2013-2022 ETH Zurich, Institute for Particle and Astrophysics and SIS
# ID.
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.

import sympy as sp


class Min(sp.Expr):
    def __init__(self, a, b):
        self.a = a
        self.b = b

    @property
    def free_symbols(self):
        return self.a.free_symbols | self.b.free_symbols

    def to_c(self, visitor):
        return "(_min({}, {}))".format(visitor.visit(self.a), visitor.visit(self.b))


class Max(Min):
    def to_c(self, visitor):
        return "(_max({}, {}))".format(visitor.visit(self.a), visitor.visit(self.b))


class isnan(sp.Expr):
    def __init__(self, a):
        self.a = a

    @property
    def free_symbols(self):
        return self.a.free_symbols

    def to_c(self, visitor):
        # from https://stackoverflow.com/questions/570669
        return "({0} != {0})".format(visitor.visit(self.a))


class hyper_2f1(sp.Expr):
    def __init__(self, a, b, c, x):
        self.a = a
        self.b = b
        self.c = c
        self.x = x

    @property
    def free_symbols(self):
        return (
            self.a.free_symbols
            | self.b.free_symbols
            | self.c.free_symbols
            | self.x.free_symbols
        )

    def to_c(self, visitor):
        # from https://stackoverflow.com/questions/570669
        return "(gsl_sf_hyperg_2F1({}, {}, {}, {}))".format(
            visitor.visit(self.a),
            visitor.visit(self.b),
            visitor.visit(self.c),
            visitor.visit(self.x),
        )



# This file is part of sympy2c.
#
# Copyright (C) 2013-2022 ETH Zurich, Institute for Particle and Astrophysics and SIS
# ID.
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.

import functools
import hashlib
import os
import platform
import sys
import time
from contextlib import contextmanager


def get_platform():
    os = platform.system()
    if os != "Darwin":
        return "-".join([platform.system().lower(), platform.machine()])

    version = platform.mac_ver()[0].rsplit(".", 1)[0]
    machine = platform.mac_ver()[-1]
    return "-".join(["macosx", version, machine])


def create_folder_if_not_exists(folder):
    if not os.path.exists(folder):
        try:
            os.makedirs(folder)
        except IOError:
            # might be triggered due to race condition
            assert os.path.exists(folder)


def align(text):
    """removes | markers and everyting left to them"""
    lines = text.split("\n") + ["", ""]
    return "\n".join(line.partition("|")[2] for line in lines)


def concat_generator_results(function):
    """code generator methods here *yield* code blocks.
    This decorator then *returns* the concatendated blocks
    as a single string
    """

    @functools.wraps(function)
    def inner(*a, **kw):
        concat = "\n".join(function(*a, **kw))
        stripped = [line.rstrip() for line in concat.split("\n")]
        return "\n".join(stripped)

    return inner


def bound_symbols(*expressions):
    def union(generator):
        return set.union(*list(generator))

    from sympy import Symbol

    all_symbols = union(expression.atoms(Symbol) for expression in expressions)
    free_symbols = union(expression.free_symbols for expression in expressions)

    return all_symbols - free_symbols


@contextmanager
def run_in_folder(folder):
    create_folder_if_not_exists(folder)
    current_folder = os.getcwd()
    try:
        os.chdir(folder)
        yield
    finally:
        os.chdir(current_folder)


@contextmanager
def timeit(task):
    print("start", task)
    started = time.time()
    yield
    print("time needed for {}: {:.2f} seconds".format(task, time.time() - started))


PRINT_INFO = os.environ.get("PRINTHASHES") is not None


class Hasher:
    def __init__(self):
        self._hasher = hashlib.md5()

    def update(self, name, what):
        self._hasher.update(what.encode("utf-8"))
        if PRINT_INFO:
            print("HASH", name, self._hasher.hexdigest(), what)

    def hexdigest(self):
        return self._hasher.hexdigest()[:8]


def base_cache_folder():
    home = os.path.expanduser("~")
    if sys.platform.startswith("darwin"):
        base_folder = os.path.join(home, "Library", "Cache")
    else:
        base_folder = os.path.join(home, "_cache")
    return base_folder


def sympy2c_cache_folder():
    fake_folder = os.environ.get("FAKE_CACHE_FOLDER")
    if fake_folder is not None:
        return fake_folder
    import numpy

    from sympy2c import __version__

    numpy_version = "_".join(numpy.__version__.split("."))
    sympy_version = __version__.replace(".", "_")
    python_version = str(sys.version_info.major) + str(sys.version_info.minor)
    return os.path.join(
        base_cache_folder(),
        "sympy2c",
        get_platform(),
        sympy_version + "__np_" + numpy_version + "_py" + python_version,
    )



# This file is part of sympy2c.
#
# Copyright (C) 2013-2022 ETH Zurich, Institute for Particle and Astrophysics and SIS
# ID.
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.

import hashlib
import pickle

from .utils import align, concat_generator_results


class Globals(object):
    def __init__(self, *variables):
        self.variables = set(variables)

    def add_variable(self, symbol):
        assert symbol not in self.variables, "variable was already set as being global"
        self.variables.add(symbol)

    def get_unique_id(self):
        data = sorted(map(str, self.variables))
        return hashlib.md5(pickle.dumps(data)).hexdigest()

    def __iter__(self):
        return iter(self.variables)

    @concat_generator_results
    def c_code(self, header_file_path):
        for variable in self.variables:
            yield "static double {};".format(variable.c_args_decl())

        yield ""
        for variable in self.variables:
            yield variable.setter_c_code()
            yield variable.getter_c_code()
        yield ""

    @concat_generator_results
    def c_header(self):
        yield """extern "C" { """
        for variable in self.variables:
            yield "    " + variable.setter_header_code()
            yield "    " + variable.getter_header_code()
        yield "}"
        yield ""

    @concat_generator_results
    def cython_code(self, header_file_path):
        if self.variables:
            yield """cdef extern from "{header_file_path}": """.format(
                header_file_path=header_file_path
            )

            for variable in self.variables:
                yield "    " + variable.setter_cython_header_code()
                yield "    " + variable.getter_cython_header_code()
            yield ""

        yield "def set_globals(**_g):"

        if self.variables:
            yield "    cdef np.ndarray[np.double_t, ndim=1] values_1d"
            for variable in self.variables:
                for line in variable.cython_globals_setter_code().split("\n"):
                    yield "    " + line

        else:
            yield "    pass"

        yield align(
            r"""
                |class _Globals:
                |    def __setattr__(self, name, value):
                |        set_globals(**{name: value})
                |    def __getattr__(self, name):
                |        return get_globals()[name]
                |    def __str__(self):
                |        lines = []
                |        max_length = max(len(k) for k in get_globals().keys())
                |        for (key, value) in sorted(get_globals().items()):
                |            lines.append(
                |                 "{} = {}".format(key.ljust(max_length, " "), value)
                |            )
                |        return "\n".join(lines)
                |    def __getstate__(self):
                |        return None
                |    def __setstate__(self, data):
                |        pass
                |globals = _Globals()
                """
        )

        yield ""

        yield "def get_globals():"
        yield "    cdef np.ndarray[np.double_t, ndim=1] values_1d"
        yield "    result = {}"

        for variable in self.variables or []:
            for line in variable.cython_globals_getter_code().split("\n"):
                yield "    " + line
        yield "    return result"
        yield ""



# This file is part of sympy2c.
#
# Copyright (C) 2013-2022 ETH Zurich, Institute for Particle and Astrophysics and SIS
# ID.
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.

import itertools
import os
import pickle
from collections import defaultdict
from textwrap import dedent
from time import time

import sympy as sp
from sympy import Identity, Matrix, cse

from .rec_inv import rec_inv
from .speedup import jacobian
from .symbol import Symbol, symbols
from .utils import create_folder_if_not_exists, sympy2c_cache_folder, timeit
from .vector import VectorElement

"""

FROM https://math.boisestate.edu/~wright/courses/m566/luexample2print.pdf


M = LUP  (P is row permutations)

# lu
for k = 1:n-1
    for i = k+1:n  % Loop over the rows in column k+1
        A(i,k) = A(i,k)/A(k,k);
        for j = k+1:n  % Loop over the columns in row i
            A(i,j) = A(i,j) - A(i,k)*A(k,j);

# forward sub

x(1) = b(1);
for j = 2:n
    x(j) = b(j) - L(j,1:j-1)*x(1:j-1);
end

# back sub

x(n) = x(n)/U(n,n);
for j = n-1:-1:1
    x(j) = ( x(j) - U(j,j+1:n)*x(j+1:n) )/U(j,j);
    end
"""


g = itertools.count(0)
xi_symbols_generator = (Symbol(f"_x_{i:03d}") for i in g)  # in itertools.count())


def partial_cse(e):
    ignore = e.atoms(sp.Symbol) - e.free_symbols
    return cse(e, xi_symbols_generator, ignore=ignore)


class Cache:
    def __init__(self, unique_id_ode):
        folder = os.path.join(sympy2c_cache_folder(), "lu_generator")
        create_folder_if_not_exists(folder)
        self._path = os.path.join(folder, unique_id_ode)
        if os.path.exists(self._path):
            with open(self._path, "rb") as fh:
                self._cache = pickle.load(fh)
        else:
            self._cache = {}

    def __contains__(self, key):
        return key in self._cache

    def lookup(self, key):
        return self._cache[key]

    def store(self, key, value):
        self._cache[key] = value
        with open(self._path, "wb") as fh:
            pickle.dump(self._cache, fh)


def setup_code_generation(
    rhs, time_var, state_vars, visitor, unique_id_ode, permutation
):
    variable_names = set(str(s) for s in rhs.atoms()) | set((str(time_var),))
    assert (
        "h" not in variable_names
    ), "you must not use the symbol h in your expressions"
    assert (
        "el0" not in variable_names
    ), "you must not use the symbol el0 in your expressions"

    for term in rhs:
        visitor.visit(term)

    t = Symbol("_t")
    rhs = rhs.subs({time_var: t})

    cache = Cache(unique_id_ode)
    N = len(rhs)

    # TODO: more caching!
    with timeit("compute jacobian and M"):
        if ("jacobian", permutation) in cache:
            print("... cache hit")
            J, h, el0, M = cache.lookup(("jacobian", permutation))
        else:
            J = jacobian(rhs, state_vars, N=4)
            h, el0 = symbols("ls0001_.h ls0001_.el0")
            M = (Identity(N) - h * el0 * J).as_mutable()
            cache.store(("jacobian", permutation), (J, h, el0, M))

    count_nz = sum(1 for i in range(M.rows) for j in range(M.rows) if M[i, j] != 0)
    print(
        "n=",
        M.rows,
        "non zeros:",
        count_nz,
        "sparsity: %.3f %%" % (100 * count_nz / M.rows**2),
    )
    return cache, J, h, el0, M, rhs


def generate_code(
    id_, rhs, state_vars, visitor, traces, splits, cache, J, h, el0, M, permutation
):
    N = len(rhs)
    if splits is None:
        splits = [N]
    splits = sorted(splits)

    with timeit("run cse"):
        key = ("M_reduced", permutation)
        if key in cache:
            print("... cache hit")
            sub_assignments, M_reduced = cache.lookup(key)
        else:
            sub_assignments, (M_reduced,) = partial_cse(M)
            cache.store(key, (sub_assignments, M_reduced))

    sizes = [
        s
        for s in [splits[0]]
        + [s1 - s0 for (s0, s1) in zip(splits, splits[1:])]
        + [N - splits[-1]]
        if s > 0
    ]
    if not traces:
        traces = {i: [list(range(s))] for i, s in enumerate(sizes)}
    assert isinstance(traces, dict)

    print("create header")
    yield from generate_header(id_, rhs, N, traces, splits)

    print("create fnorm_")
    yield from generate_fnorm_(id_, J, state_vars, visitor.visit)

    yield from create_split_lu(
        id_,
        sub_assignments,
        M_reduced,
        traces,
        state_vars,
        visitor.visit,
        splits,
        cache,
        permutation,
    )

    print("create rhs")
    yield from generate_code_rhs(id_, rhs, state_vars, visitor.visit)


def _split(M, n):
    A = M[:n, :n]
    B = M[:n, n:]
    C = M[n:, :n]
    D = M[n:, n:]

    s = time()
    if D.shape[0]:  # messages for corner cases might be confusing
        fill = (
            sum(1 for i in range(D.shape[0]) for j in range(D.shape[1]) if D[i, j] != 0)
            / D.shape[0]
            / D.shape[1]
        )
        print(f"start inverting D of shape {D.shape}, fill_degree={fill*100:.0f}%")
    xii, Dinv = rec_inv(D)  # D.inv()
    needed = time() - s
    if D.shape[0]:  # messages for corner cases might be confusing
        print(f"inverting D of shape {D.shape} needed {needed:.1f}[s]")
    U = B * Dinv
    R = A - U * C
    return xii, R, U, C, D


def compute_splits(M, splits):
    R = M
    Di_Ci = []
    split_sub_assignments = []
    for split in reversed(splits):
        xii, R, U, C, D = _split(R, split)
        split_sub_assignments.extend(xii)
        if D:
            Di_Ci.append((D, C, split))
    return split_sub_assignments, R, U, Di_Ci


def create_split_lu(
    id_,
    sub_assignments,
    M_reduced,
    all_traces,
    state_vars,
    expression_to_c,
    splits,
    cache,
    permutation,
):
    yield from _create_decl_lu(id_, 0)
    for trace_idx, trace in enumerate(all_traces[0]):
        yield from _create_decl_split_lu_0(id_, trace_idx)

    for level, traces in all_traces.items():
        if level == 0:
            yield from _create_decl_split_lu_0(id_, trace_idx)
        else:
            yield from _create_decl_lu(id_, level)
            for trace_idx in range(len(traces)):
                yield from _create_decl_split_lu_i(id_, level, trace_idx)

    splits = tuple(splits)
    with timeit("compute schur complement matrices"):
        key = (splits, permutation)
        if key in cache:
            print("... cache hit")
            R, U, Di_Ci, sub_assignments, split_sub_assignments = cache.lookup(key)

        else:
            split_sub_assignments, R, U, Di_Ci = compute_splits(M_reduced, splits)
            cache.store(
                key,
                (R, U, Di_Ci, sub_assignments, split_sub_assignments),
            )

    traces = all_traces[0]

    yield from create_lu(
        id_,
        0,
        sub_assignments + split_sub_assignments,
        R,
        traces,
        state_vars,
        expression_to_c,
    )
    yield from create_lu_fallback_0(
        id_,
        sub_assignments + split_sub_assignments,
        U,
        traces,
        state_vars,
        expression_to_c,
    )

    for trace_idx in range(len(traces)):
        yield from _create_split_lu_0(
            id_,
            R,
            U,
            traces,
            trace_idx,
            sub_assignments + split_sub_assignments,
            state_vars,
            expression_to_c,
        )

    for level, (Di, Ci, split) in enumerate(reversed(Di_Ci), 1):
        traces = all_traces[level]

        yield from create_lu(
            id_,
            level,
            sub_assignments + split_sub_assignments,
            Di,
            traces,
            state_vars,
            expression_to_c,
        )
        yield from create_lu_fallback_i(
            id_,
            level,
            sub_assignments + split_sub_assignments,
            Ci,
            traces,
            split,
            state_vars,
            expression_to_c,
        )
        for trace_idx in range(len(traces)):
            yield from _create_split_lu_i(
                id_,
                level,
                Di,
                Ci,
                traces,
                trace_idx,
                split,
                sub_assignments + split_sub_assignments,
                state_vars,
                expression_to_c,
            )

    yield from generate_solve_split(id_, all_traces, splits, M_reduced.shape[0])


def generate_solve_split(id_, all_traces, splits, N):
    n_traces = len(all_traces[0])

    tlens = ", ".join(
        str(len(all_traces[level]) + 1) for level in range(len(all_traces))
    )

    yield dedent(
        rf"""

    std::vector<std::vector<unsigned int> > _call_counts_{id_};
    static unsigned int  _next_solver_{id_}[{len(splits) + 1}] = {{0}};

    extern "C" void _reset_next_solver_{id_}() {{
        for (int i=0; i <= {len(splits)}; i++)
            _next_solver_{id_}[i] = 0;  // solve_lu
    }}
    extern "C" void _solve_fast_{id_}(double * _t, double *_x, double *_y, int *ier) {{

        // lsoda assumes that we have _b and _x in the same memory location
        // but for our solvers this results in conflicts, thus we need to
        // separate _b and _x:

        std::vector<unsigned int> _cc;
        unsigned int _n[] = {{ {tlens} }};
        if (_call_counts_{id_}.size() == 0) {{
            for (int _ii=0; _ii < {len(all_traces)}; ++_ii) {{
                    _cc.clear();
                    _cc.insert(_cc.begin(), _n[_ii], 0);
                    _call_counts_{id_}.push_back(_cc);
                }}
        }}
        double _b[{N}];
        double _dummy[{N}];
        for (unsigned int _i = 0; _i < {N}; _i++) _b[_i] = _x[_i];

    """
    )

    for level, traces in all_traces.items():
        yield f"    int solver_{level} = _next_solver_{id_}[{level}];"

    yield ""
    yield f"    int _tried_0[{n_traces}] = {{ 0 }};"
    yield f"    if (__enable_fast_solver_{id_} == 0) solver_0 = {n_traces};"

    for level, traces in all_traces.items():
        yield f"    _call_counts_{id_}.at({level}).at(solver_{level})++;"

    for j in range(len(all_traces[0])):
        yield f"    if (solver_0 == {j})"
        yield (
            f"        _next_solver_{id_}[0] = solve_{id_}__0_{j}(*_t, _b, _x, _x, _y,"
            f"_tried_0, __enable_sparse_lu_solver_{id_});"
        )

    yield f"    if (solver_0 == {j + 1})"
    yield (
        f"        _next_solver_{id_}[0] = solve_lu_{id_}__0_f(*_t, _b, _x, _x, _y,"
        f"_tried_0, __enable_sparse_lu_solver_{id_});"
    )
    yield ""
    offset = len(all_traces[0]) + 1

    for level, split in enumerate(splits, 1):
        if split == N:
            continue
        n_traces = len(all_traces[level])
        yield f"    int _tried_{level}[{n_traces}] = {{ 0 }};"
        yield f"    if (__enable_fast_solver_{id_} == 0)"
        yield f"        solver_{level} = {len(all_traces[level])};"
        for j in range(len(all_traces[level])):
            yield f"    if (solver_{level} == {j})"
            yield (
                f"        _next_solver_{id_}[{level}] ="
                f" solve_{id_}__{level}_{j}(*_t, _b, _x, _x + {split}, _y,"
                f" _tried_{level}, __enable_sparse_lu_solver_{id_});"
            )

        yield f"    if (solver_{level} == {j + 1})"
        yield (
            f"        _next_solver_{id_}[{level}] ="
            f" solve_lu_{id_}__{level}_f(*_t, _b, _x, _x + {split}, _y,"
            f" _tried_{level}, __enable_sparse_lu_solver_{id_});"
        )
        offset *= n_traces + 1

    yield "}"


def _create_decl_split_lu_0(id_, trace_idx):
    yield (
        f'extern "C" int solve_{id_}__0_{trace_idx}(double _t,'
        f" double *__restrict__ __b, double * _x0, "
        f" double *__restrict__ _x, double *__restrict__ _y, int * _tried,"
        " int _consider_sparsity);"
    )


def _create_decl_split_lu_i(id_, level, trace_idx):
    yield (
        f'extern "C" int solve_{id_}__{level}_{trace_idx}(double _t, '
        f"double *__restrict__ __b, double *_x0, "
        f"double *__restrict__ _x, double *__restrict__ _y, int * _tried,"
        " int _consider_sparsity);"
    )


def _create_split_lu_0(
    id_, R, U, traces, trace_idx, sub_assignments, state_vars, expression_to_c
):
    n0, n1 = U.shape
    trace = traces[trace_idx]

    yield (
        f'extern "C" int solve_{id_}__0_{trace_idx}(double _t,'
        f" double *__restrict__ __b, double * _x0, "
        f" double *__restrict__ _x, double *__restrict__ _y, int * _tried, "
        " int _consider_sparsity) {"
    )

    yield ""
    yield from init_state_vars_from_y(state_vars, indent="    ")
    yield from create_xi_varables(sub_assignments, expression_to_c)

    yield ""
    yield f"    double _b[{n0}];"

    for i in range(n0):
        yield f"    _b[{i}] = __b[{i}];"
        for j in range(n1):
            xij = U[i, j]
            if xij != 0:
                xij_expr = expression_to_c(xij)
                yield f"    _b[{i}] -= __b[{n0} + {j}] * ({xij_expr});"

    yield ""
    yield "    double _v_max, _temp;"
    yield "    int _j_max;"
    yield f"    _tried[{trace_idx}] = 1;"
    yield ""

    yield from create_sparse_lu(id_, 0, R, expression_to_c, trace, traces)

    yield f"    return {trace_idx};"
    yield "}"


def _create_split_lu_i(
    id_,
    level,
    D,
    C,
    traces,
    trace_idx,
    b_offset,
    sub_assignments,
    state_vars,
    expression_to_c,
):
    n1 = D.shape[0]
    n0 = C.shape[1]

    trace = traces[trace_idx]

    yield (
        f'extern "C" int solve_{id_}__{level}_{trace_idx}(double _t, '
        f"double *__restrict__ __b, double *_x0, "
        f"double *__restrict__ _x, double *__restrict__ _y, int * _tried,"
        " int _consider_sparsity) {"
    )
    yield ""
    yield from init_state_vars_from_y(state_vars, indent="    ")
    yield from create_xi_varables(sub_assignments, expression_to_c)

    yield ""
    yield f"    double _b[{n1}];"

    yield ""
    yield "    double _v_max, _temp;"
    yield "    int _j_max;"
    yield f"    _tried[{trace_idx}] = 1;"
    yield ""

    for i in range(n1):
        yield f"    _b[{i}] = __b[{i} + {b_offset}];"
        for j in range(n0):
            if C[i, j] != 0:
                cij_expr = expression_to_c(C[i, j])
                yield f"    _b[{i}] -= _x0[{j}] * ({cij_expr});"

    yield from create_sparse_lu(id_, level, D, expression_to_c, trace, traces)
    yield f"    return {trace_idx};"
    yield "}"


def _create_decl_lu(id_, level):
    yield (
        'extern "C" '
        f"int solve_lu_{id_}__{level}(double _t, double* _b, double *_x0, double* _x,"
        " double* _y, int* _tried, int _consider_sparsity);"
    )


def create_lu_fallback_0(id_, sub_assignments, U, traces, state_vars, expression_to_c):
    n0, n1 = U.shape

    yield dedent(
        f"""
    extern "C"
    int solve_lu_{id_}__0_f(double _t, double* __b, double *_x0, double* _x, double* _y,
                            int* _tried, int _consider_sparsity)
    {{
    """
    )
    yield from init_state_vars_from_y(state_vars, indent="    ")
    yield from create_xi_varables(sub_assignments, expression_to_c)
    yield ""
    yield f"    double _b[{n0}];"

    for i in range(n0):
        yield f"    _b[{i}] = __b[{i}];"
        for j in range(n1):
            xij = U[i, j]
            if xij != 0:
                xij_expr = expression_to_c(xij)
                yield f"    _b[{i}] -= __b[{n0} + {j}] * ({xij_expr});"

    yield (
        f"    return solve_lu_{id_}__0(_t, _b, _x0, _x, _y, _tried,"
        " _consider_sparsity);"
    )
    yield "}"


def create_lu_fallback_i(
    id_, level, sub_assignments, C, traces, b_offset, state_vars, expression_to_c
):
    n0, n1 = C.shape

    yield dedent(
        f"""
    extern "C"
    int solve_lu_{id_}__{level}_f(double _t, double* __b, double * _x0, double* _x,
                                  double* _y, int* _tried, int _consider_sparsity)
    {{
    """
    )
    yield from init_state_vars_from_y(state_vars, indent="    ")
    yield from create_xi_varables(sub_assignments, expression_to_c)
    yield ""
    yield f"    double _b[{n0}];"

    for i in range(n0):
        yield f"    _b[{i}] = __b[{i} + {b_offset}];"
        for j in range(n1):
            if C[i, j] != 0:
                cij_expr = expression_to_c(C[i, j])
                yield f"    _b[{i}] -= _x0[{j}] * ({cij_expr});"

    yield (
        f"    return solve_lu_{id_}__{level}(_t, _b, _x0, _x, _y, _tried, "
        "_consider_sparsity);"
    )
    yield "}"


def create_lu(
    id_,
    level,
    sub_assignments,
    M_reduced,
    traces,
    state_vars,
    expression_to_c,
):
    N = M_reduced.rows
    n = len(traces)

    lines = [
        f"double {xi} = {expression_to_c(expr)};" for (xi, expr) in sub_assignments
    ]
    sub_assignment_code = "\n        ".join(lines)

    M_init_code = "\n        ".join(
        f"_M({i}, {j}) = {expression_to_c(M_reduced[i, j])};"
        for i in range(N)
        for j in range(N)
        if M_reduced[i, j] != 0
    )

    col_end = [max(i for i in range(N) if M_reduced[i, j] != 0) for j in range(N)]
    row_end = [max(j for j in range(N) if M_reduced[i, j] != 0) for i in range(N)]

    # col_end = row_end = [N - 1] * N

    row_end_str = ", ".join(map(str, row_end))
    col_end_str = ", ".join(map(str, col_end))

    init_state_vars_code = "\n        ".join(init_state_vars_from_y(state_vars))

    code = f"""
    // FROM https://math.boisestate.edu/~wright/courses/m566/luexample2print.pdf

    #undef _M
    #define _M(i, j) __M[i + j * {N}]

    #define PRINT_TRACES 0

    extern "C"
    int solve_lu_{id_}__{level}(double _t, double* _b, double *_x0, double* _x,
                                double* _y, int* _tried, int  _consider_sparsity)
    {{
        // printf("enter solve_lu_{id_}\\n");

        reset_timer();

        int    _i, _j, _k, _i_max, _count, _pi, _pk;
        double _v_max, _temp;
        double __M[{N} * {N}] = {{0}};

        int _perm[{N}];
        int _trace[{N}];

        int _row_ends[{N}] = {{ {row_end_str} }};
        int _col_ends[{N}] = {{ {col_end_str} }};

        for (_i = 0; _i < {N}; _i++) _perm[_i] = _trace[_i] = _i;

        {init_state_vars_code}
        {sub_assignment_code}
        {M_init_code}

        if (0)
            for (int _i=0; _i < {N}; _i++) {{
                for (int _j=0; _j < {N}; _j++)
                    printf("%+.1e  ", _M(_i, _j));
                printf("\\n");
            }}


        double _v;

        int _col_end_k, _row_end_k, _row_end_i;

        _col_end_k = _row_end_k = _row_end_i = {N - 1};

        for (_k = 0; _k < {N-1}; _k++) {{

            _v_max = fabs(_M(_perm[_k], _k));

            _i_max = _k;

             if (_consider_sparsity)
                _col_end_k = _col_ends[_k];

            for (_j = _k + 1; _j <= _col_end_k; _j++) {{
                if ((_temp = fabs(_M(_perm[_j], _k))) > _v_max * SEC_FACTOR)
                {{
                    _v_max = _temp;
                    _i_max = _j;
                }}
            }}
            if (_v_max == 0.0) return -2;
            _j = _perm[_k];
            _perm[_k] = _perm[_i_max];
            _perm[_i_max] = _j;
            // printf("SWAP %d %d\\n", _k, _i_max);

            _trace[_k] = _i_max;

            _pk = _perm[_k];
            if (_consider_sparsity)
                _row_end_k = _row_ends[_pk];

            for (_i = _k + 1; _i <= _col_end_k; _i++) {{

                _pi = _perm[_i];

                _M(_pi, _k) /= _M(_pk, _k);


                if (_consider_sparsity) {{
                    _row_end_k = _row_ends[_pk];
                    _row_end_i = _row_ends[_pi];

                    for (_j = _k + 1; _j <= _row_end_k; _j++) {{
                        _M(_pi, _j) -= _M(_pi, _k) * _M(_pk, _j);
                        if (_M(_pi, _j) != 0) {{
                            if (_j > _row_end_i) {{
                                if (0) printf("extend row_ends %d from %d to %d\\n",
                                              _pi, _row_end_i, _j);
                                _row_end_i = _row_ends[_pi] = _j;
                            }}
                            if (_pi > _col_ends[_j]) {{
                                if (0) printf("extend col_ends %d from %d to %d\\n",
                                              _j, _col_ends[_j], _pi);
                                _col_ends[_j] = _pi;
                            }}
                        }} else {{
                            if (_j == _row_end_i)
                            _row_end_i = _row_ends[_pi] = _k - 1;
                            if (_pi == _col_ends[_j]) {{
                                if (0) printf("shrink col_ends %d from %d to %d\\n",
                                              _j, _col_ends[_j], _pi - 1);
                                _col_ends[_j]--;
                                }}
                        }}
                    }}
                }}
                else
                    for (_j = _k + 1; _j <= _row_end_k; _j++)
                        _M(_pi, _j) -= _M(_pi, _k) * _M(_pk, _j);
            }}

        }}
        if (0)
            for (int _i=0; _i < {N}; _i++) {{
                for (int _j=0; _j < {N}; _j++)
                    printf("%+.1e  ", _M(_i, _j));
                printf("\\n");
            }}

        if (0)
            for (int _j=0; _j < {N}; _j++) {{
                int _ce = 0;
                for (int _i=0; _i < {N}; _i++) {{
                    if(_M(_i, _j) !=0)
                        _ce = _i;
                }}
                printf("col_end %d: %d %d\\n", _j, _col_ends[_j], _ce);

            }}
        if (0)
            for (int _i=0; _i < {N}; _i++) {{
                int _re = 0;
                for (int _j=0; _j < {N}; _j++)
                    if(_M(_i, _j) !=0)
                        _re = _j;
                printf("row_end %d: %d %d\\n", _i, _row_ends[_i], _re);
            }}

        _v_max = fabs(_M(_perm[_k], _k));
        if (_v_max == 0.0) return -2;

        // forward substitution

        _x[0] = _b[_perm[0]];
        for (_j = 1; _j < {N}; _j++) {{
            _x[_j] = _b[_perm[_j]];
            for (_k = 0; _k < _j; _k++)
                _x[_j] -= _M(_perm[_j], _k) * _x[_k];
        }}

        // backward hubstitution

        _x[{N-1}] = _x[{N-1}] / _M(_perm[{N - 1}], {N - 1});
        for (_j = {N-2}; _j >=0; _j--) {{
            for (_k = _j + 1; _k < {N}; _k++)
                _x[_j] -= _M(_perm[_j], _k) * _x[_k];
            _x[_j] /= _M(_perm[_j], _j);
        }}

        _total_time_{id_}_lup_solver += time_delta();

        for (_j = 0; _j < {n}; _j++) {{
            _count = 0;
            for (_i = 0;  _i < {N}; _i++) {{
                if (_trace[_i] != _traces_{id_}__{level}[_j][_i])
                    break;
                _count += 1;
            }}
            if (_count == {N}) {{
                // printf("leave solve_lu_{id_} with trace %d\\n", _j);
                return _j;
                }}
        }}

        // register new trace

        std::vector<int> _new_trace;
        _new_trace.reserve({N});
        for (_i = 0;  _i < {N}; _i++)
            _new_trace.push_back(_trace[_i]);

        int _before = _new_traces["{id_}"][{level}].size();

        _new_traces["{id_}"][{level}].insert(_new_trace);

        int _after = _new_traces["{id_}"][{level}].size();

        if (PRINT_TRACES)
            if (_before != _after) {{
                printf("new trace %d\\n", _after);
                for (_j = 0; _j < {N}; _j++) {{
                    for (_i = 0; _i < {N}; _i++)
                        printf("  %+e,", _M(_i, _j));
                    printf("\\n");
                }}
                printf("rhs: ");
                for (_i = 0; _i < {N}; _i++) printf(" %e, ", _b[_i]);
                printf("\\nsolution: ");
                for (_i = 0; _i < {N}; _i++) printf(" %e, ", _x[_i]);
                printf("\\n");
            }}

        // printf("leave solve_lu_{id_} with new trace\\n");
        return {len(traces)};
    }}
    """
    yield dedent(code)


def init_state_vars_from_y(state_vars, indent=""):
    vector_elements = defaultdict(list)
    for i, state_var in enumerate(state_vars):
        if isinstance(state_var, VectorElement):
            vector_elements[state_var.vector_name].append((state_var.index, i))
        else:
            yield f"{indent}double {state_var} = _y[{i}];"

    for vector_name, used_elements in vector_elements.items():
        array_size = max(used_elements)[0] + 1
        yield f"{indent}double {vector_name}[{array_size}];"
        for vec_index, y_index in used_elements:
            yield f"{indent}{vector_name}[{vec_index}] = _y[{y_index}];"


def generate_fnorm_(id_, J, state_vars, expression_to_c):
    N = J.rows

    # c function name ends with "_" because of fortran linking. The modified
    # lsoda calls fnorm from Fortran wich is the same as fnorm_ in C.

    code = f"""
    extern "C" double _fnorm_{id_}(double *__t, double *_ewt, double *_y) {{
        // c function name ends with "_" because of fortran linking. The modified
        // lsoda calls fnorm from Fortran wich is the same as fnorm_ in C.
        double _fnorm = 0.0, _row_norm;
        """
    yield dedent(code)

    yield from init_state_vars_from_y(state_vars, indent="    ")

    yield ""
    yield "    double _t = *__t;"
    yield ""

    sub_assignments, (new_expressions,) = partial_cse(J)

    for xi, expr in sub_assignments:
        c_expr = expression_to_c(expr)
        yield f"    double {xi} = {c_expr};"

    for i in range(N):
        yield "    _row_norm = 0.0f;"
        for j in range(N):
            if J[i, j] != 0:
                expression = expression_to_c(new_expressions[i, j])
                yield f"    _row_norm += fabs({expression} / _ewt[{j}]);"
        yield f"    _row_norm *= _ewt[{i}];"
        yield "    if (_row_norm > _fnorm) _fnorm = _row_norm;"
    # yield r'   printf("fnorm = %e\n", _fnorm);'
    yield "    return _fnorm;"
    yield "}"


def generate_header(id_, rhs, N, traces, splits):
    yield ""
    yield f"#define N_{id_} {N}"
    yield ""
    yield f"static double _total_time_{id_}_fast_solver;"
    yield f"static double _total_time_{id_}_lup_solver;"
    yield f"static int __enable_fast_solver_{id_} = 1;"
    yield f"static int __enable_sparse_lu_solver_{id_} = 1;"
    yield ""
    yield f'extern "C" void _reset_timers_{id_}() {{'
    yield f"        _total_time_{id_}_fast_solver = 0.0;"
    yield f"        _total_time_{id_}_lup_solver = 0.0;"
    yield "}"
    yield ""
    yield f'extern "C" void _enable_fast_solver_{id_}(int enable) {{'
    yield f"        __enable_fast_solver_{id_} = enable;"
    yield "}"
    yield f'extern "C" void _enable_sparse_lu_solver_{id_}(int enable) {{'
    yield f"        __enable_sparse_lu_solver_{id_} = enable;"
    yield "}"
    yield ""
    yield f'extern "C" double _get_timer_{id_}_fast() {{'
    yield f"        return _total_time_{id_}_fast_solver;"
    yield "}"
    yield ""
    yield f'extern "C" double _get_timer_{id_}_lup() {{'
    yield f"        return _total_time_{id_}_lup_solver;"
    yield "}"

    for level, level_traces in traces.items():
        # tri = [ (t0, ...tl), (t0, .. tl), ...]
        yield f"std::vector<std::vector<int>> _traces_{id_}__{level} = {{"
        for trace in level_traces:
            yield "        {{ {} }},".format(", ".join(map(str, trace)))
        yield "};"

    yield (
        f'extern "C" int solve_lu_{id_}(double, double*, double *, double*, double*,'
        " int*, int);"
    )
    yield ""


def create_check(id_, M, expression_to_c, state_vars):
    n = M.shape[0]
    yield f"void _check_{id_}(double _t, double *_x, double *_b, double *_y) {{"

    yield from init_state_vars_from_y(state_vars, indent="    ")
    yield "     double _sum;"

    for i in range(n):
        yield "    _sum = 0.0;"
        for j in range(n):
            if M[i, j] != 0:
                mij_expr = expression_to_c(M[i, j])
                yield f"    _sum += _x[{j}] * {mij_expr};"
        yield rf"""    printf("r{i}= %e ", (_b[{i}] - _sum) / _b[{i}]);"""
    yield r"""    printf("\n");"""
    yield "}"


def create_sparse_lu(id_, level, M, expression_to_c, trace, traces):
    N = M.rows
    Me = [[expression_to_c(M[i, j]) for j in range(N)] for i in range(N)]

    perm = list(range(N))

    declared = set()

    yield "    reset_timer();"

    for i in range(N):
        for j in range(N):
            if M[i, j]:
                yield f"    double _m_{i}_{j} = {Me[i][j]};"
                declared.add((i, j))

    for k in range(N - 1):
        pk = perm[k]
        if (pk, k) not in declared:
            yield f"    double _m_{pk}_{k} = {Me[pk][k]};"
            declared.add((pk, k))
        yield f"    _v_max = fabs(_m_{pk}_{k});"

        yield f"    _j_max = {k};"
        # yield f"""printf("k={k}, _v_max=%f\\n", _v_max);"""

        for j in range(k + 1, N):
            pj = perm[j]
            if (pj, k) in declared:
                yield f"    if ((_temp = fabs(_m_{pj}_{k})) > _v_max * SEC_FACTOR)"
                yield f"       {{ _j_max = {j}; _v_max = _temp; }}"

        yield f"    if (_j_max != {trace[k]}) {{"
        yield f'        // printf("try trace switch for k={k} j_max=%d\\n", _j_max);'

        for nt, t in enumerate(traces):
            if any(t[i] != trace[i] for i in range(k)):
                continue

            if t[k] == trace[k]:
                continue

            yield f"        if (_j_max == {t[k]} && !_tried[{nt}]) {{"
            yield (
                f"            return solve_{id_}__{level}_{nt}(_t, __b, _x, _x, _y,"
                " _tried, _consider_sparsity);"
            )
            yield "        }"
        yield (
            f"        return solve_lu_{id_}__{level}(_t, _b, _x, _x, _y, _tried,"
            " _consider_sparsity);"
        )

        yield "    }"
        yield ""

        perm[k], perm[trace[k]] = perm[trace[k]], perm[k]

        pk = perm[k]

        for i in range(k + 1, N):
            pi = perm[i]

            if (pi, k) in declared:
                if (pi, k) not in declared:
                    yield f"    double _m_{pi}_{k} = {Me[pi][k]};"
                    declared.add((pi, k))

                if (pk, k) not in declared:
                    yield f"    double _m_{pk}_{k} = {Me[pk][k]};"
                    declared.add((pk, k))
                yield f"    _m_{pi}_{k} /= _m_{pk}_{k};"

            for j in range(k + 1, N):
                if (pi, k) in declared and (
                    pk,
                    j,
                ) in declared:  # M[pi, k] and M[pk, j]:
                    if (pi, j) not in declared:
                        yield f"    double _m_{pi}_{j} = {Me[pi][j]};"
                        declared.add((pi, j))
                    if (pi, k) not in declared:
                        yield f"    double _m_{pi}_{k} = {Me[pi][k]};"
                        declared.add((pi, k))
                    if (pk, j) not in declared:
                        yield f"    double _m_{pk}_{j} = {Me[pk][j]};"
                        declared.add((pk, j))
                    yield f"    _m_{pi}_{j} -= _m_{pi}_{k} * _m_{pk}_{j};"

        yield ""

    # forward substitution
    yield "    // forward substution"
    yield ""
    yield f"    _x[0] = _b[{perm[0]}];"
    for j in range(1, N):
        pj = perm[j]
        yield f"    _x[{j}] = _b[{pj}];"
        for k in range(j):
            if (pj, k) in declared:
                yield f"    _x[{j}] -= _m_{pj}_{k} * _x[{k}];"

    # backwards substitution
    yield "    // backwards substution"
    yield ""

    if (perm[N - 1], N - 1) in declared:
        yield f"    _x[{N - 1}] /= _m_{perm[N - 1]}_{N-1};"

    for j in range(N - 2, -1, -1):
        pj = perm[j]
        for k in range(j + 1, N):
            if (pj, k) in declared:
                yield f"    _x[{j}] -= _m_{pj}_{k} * _x[{k}];"

        if (pj, j) in declared:
            yield f"    _x[{j}] /= _m_{pj}_{j};"

    yield ""
    yield f"    _total_time_{id_}_fast_solver += time_delta();"
    yield ""


def create_xi_varables(xi, expression_to_c):
    for v, t in xi:
        t = expression_to_c(t)
        yield f"    double {v} = {t};"


def generate_code_rhs(id_, rhs, state_vars, expression_to_c):
    sub_assignments, (expressions,) = partial_cse(rhs)

    init_xi = "\n        ".join(
        f"double {xii} = {expression_to_c(e)};" for (xii, e) in sub_assignments
    )

    set_ydot = "\n        ".join(
        f"_ydot[{i}] = {expression_to_c(e)};" for (i, e) in enumerate(expressions)
    )

    init_state_vars_code = "\n        ".join(init_state_vars_from_y(state_vars))

    code = f"""
    extern "C" void _rhs_{id_}(long int *_n, double *__t, double *__restrict__ _y,
                               double *__restrict__ _ydot) {{

        double _t = *__t;

        // printf("%.15e   %.15e %.15e %.15e\\n", _t, _y[0], _y[1], _y[2]);

        {init_state_vars_code}
        {init_xi}
        {set_ydot}

        // printf("rhs:  t=%.15e ydot= %.15e %.15e %.15e\\n\\n",
        //        _t, _ydot[0], _ydot[1], _ydot[2]);
    }}
    """

    yield dedent(code)


def argsort(items):
    s = sorted(zip(items, itertools.count()))
    return list(zip(*s))[1]


def main():
    a, b, c, d, e, f = symbols("a b c d e f")

    M = Matrix(
        [
            [a, b, 0, 0, 0, 1],
            [b, a, c, 0, 0, 0],
            [0, c, a, d, 0, 0],
            [0, 0, d, a, e, 0],
            [0, 0, 0, e, a, f],
            [0, 0, 0, 0, f, a],
        ]
    )

    g, h, i = symbols("g H i")

    M = Matrix([[a, 0, 0], [b, c, 0], [d, e, f]])

    M = Matrix([[a, 0, 0], [b, c, 0], [d, e, f]])
    M = Matrix([[a, b, c], [0, d, e], [0, 0, f]])

    M = Matrix([[a, b, c, d], [b, c, d, e], [c, d, e, f], [d, e, f, g]])

    M = Matrix([[a, b, c], [0, d, e], [0, 0, f]])

    M = Matrix(
        [
            [a, b, 0, 0, 0, 0, 0],
            [b, a, c, 0, 0, 0, 0],
            [0, c, a, d, 0, 0, 0],
            [0, 0, d, a, e, 0, 0],
            [0, 0, 0, e, a, f, 0],
            [h, 0, 0, 0, f, a, f],
            [0, 0, 0, 0, i, 1, g],
        ]
    )
    M = Matrix([[a, 0, 0], [b, c, 0], [d, e, f]])
    M = Matrix([[a, b, c], [d, e, f], [g, h, i]])

    N = 3
    M = 20 * Identity(N).as_mutable()

    syms = symbols("a b c d e f g i j")

    for i in range(N - 1):
        M[i, i + 1] = syms[i % len(syms)]
        M[i + 1, i] = syms[i % len(syms)]

    M[0, N - 1] = a
    for i in range(N):
        M[N - 1, i] = b

    y0, y1, y2 = symbols("_y[0] _y[1] _y[2]")
    rhs = Matrix(
        [
            -0.04 * y0 + 1e4 * y1 * y2,
            0.04 * y0 - 1e4 * y1 * y2 - 3e7 * y1**2,
            3e7 * y1**2,
        ]
    )

    state_vars = [y0, y1, y2]

    traces = [[0, 1, 2]]

    generate_code(rhs, Symbol("t"), state_vars, traces)



# This file is part of sympy2c.
#
# Copyright (C) 2013-2022 ETH Zurich, Institute for Particle and Astrophysics and SIS
# ID.
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.

import abc


class WrapperBase(abc.ABC):
    @abc.abstractmethod
    def c_header(self):
        ...

    @abc.abstractmethod
    def c_code(self, header_file_path):
        ...

    @abc.abstractmethod
    def cython_code(self, header_file_path):
        ...

    @abc.abstractmethod
    def determine_required_extra_wrappers(self):
        ...

    @abc.abstractmethod
    def setup_code_generation(self):
        ...



# This file is part of sympy2c.
#
# Copyright (C) 2013-2022 ETH Zurich, Institute for Particle and Astrophysics and SIS
# ID.
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.

import hashlib

import sympy as sp
from sympy.core.containers import Tuple
from sympy.matrices.immutable import ImmutableDenseMatrix

from .utils import align, bound_symbols, concat_generator_results
from .vector import Vector, VectorElement
from .wrapper import WrapperBase

_hash_cache = {}


def expression_hash(e):
    return hashlib.md5(str(e.sort_key()).encode("utf-8")).digest()


class Alias:
    def __init__(self, name, *symbols):
        assert isinstance(name, str)
        assert all(isinstance(symbol, sp.Symbol) for symbol in symbols)
        assert not name.startswith(
            "_"
        ), "you must not use underscore as first character"
        self.name = name
        self.symbols = symbols

    @property
    def as_vector(self):
        return Vector(self.name, len(self.symbols))

    def __str__(self):
        return "Alias({}={})".format(self.name, self.symbols)


F_with_three_args = sp.Function("Function", nargs=3, real=True)


class InternalFunction(F_with_three_args):
    def __new__(clz, name, expressions, *args):
        # fix after unpickling
        if len(args) == 1 and isinstance(args[0], Tuple):
            args = args[0]

        if not isinstance(expressions, (list, tuple)):
            expressions = [expressions]

        aliases = [a for a in args if isinstance(a, Alias)]
        args = [a.as_vector if isinstance(a, Alias) else a for a in args]

        instance = F_with_three_args.__new__(
            clz, name, sp.Matrix(expressions), tuple(args)
        )

        instance.aliases = aliases
        instance._unique_id = None
        return instance

    def __getstate__(self):
        # we need default impl, else aliases handling will not work
        return self.__dict__

    def __setstate__(self, dd):
        # we need default impl, else aliases handling will not work
        self.__dict__.update(dd)

    def get_unique_id(self):
        if self._unique_id is None:
            md5 = hashlib.md5()
            md5.update(str(self._name).encode("utf-8"))
            for e in self.expressions:
                md5.update(expression_hash(e))
            md5.update(str(self.arguments).encode("utf-8"))
            self._unique_id = md5.hexdigest()
        return self._unique_id

    @property
    def free_symbols(self):
        return set(self.args[1].free_symbols) - set(self.args[1])

    @property
    def _name(self):
        return self.args[0]

    @property
    def expressions(self):
        return list(self.args[1])

    @property
    def arguments(self):
        return self.args[2]

    def __call__(self, *args):
        assert len(args) == len(self.arguments)
        return self.args[1].subs({in_: out for in_, out in zip(self.arguments, args)})


class Function(InternalFunction):
    pass


def _signature(expression, max_len=100):
    result = str(expression)
    if len(result) > max_len:
        result = result[: max_len - 3] + "..."
    return result


def function_wrapper_factory(function, globals_, visitor):
    expose_to_python = isinstance(function, Function)
    if len(function.expressions) == 1:
        return RealValuedFunctionWrapper(function, globals_, visitor, expose_to_python)
    else:
        return VectorValuedFunctionWrapper(
            function, globals_, visitor, expose_to_python
        )


class _FunctionWrapperBase(WrapperBase):
    def __init__(self, function, globals_, visitor, expose_to_python):
        assert set(function.arguments) & set(globals_.variables) == set(())
        self.name = function._name
        self.expressions = list(function.expressions)
        self.variables = function.arguments
        self.aliases = function.aliases

        self.function = function

        self.globals_ = globals_
        self.visitor = visitor
        self.expose_to_python = expose_to_python

    def determine_required_extra_wrappers(self):
        pass

    def setup_code_generation(self):
        self.in_variables = [
            var[0] if isinstance(var, tuple) else var for var in self.variables
        ]

        if 0:
            self.args = ", ".join(
                "double {}".format(var.c_args_decl()) for var in self.in_variables
            )

        if 0:
            self.cython_args = ", ".join(
                "{}".format(var.cython_decl()) for var in self.in_variables
            )

        self.c_args_decl = ", ".join(
            ["{}".format(var.c_decl()) for var in self.in_variables]
            + ["double * __result"]
        )

        self.ctypes_args = ", ".join(
            "{}".format(var.ctypes_decl()) for var in self.in_variables
        )

        self.numba_type_decl = ", ".join(
            "{}".format(var.numba_type_decl()) for var in self.in_variables
        )

        if 1:
            self.ufunc_args = ", ".join(
                "double[{}] {}".format(":, :" if var.is_vector else ":", var._name())
                for var in self.in_variables
            )

        common_expressions, simplified_expressions = sp.cse(
            self.expressions, ignore=list(bound_symbols(*self.expressions))
        )

        # this visit calls can create integrand and integral functions which are avail
        # at self.visitor.extra_wrappers after self.visitor.visit finished:
        self.common_expressions = [
            (xi, self.visitor.visit(e)) for (xi, e) in common_expressions
        ]

        self.simplified_expressions = [
            self.visitor.visit(e) for e in simplified_expressions
        ]

    def get_unique_id(self):
        md5 = hashlib.md5()
        md5.update(self.function.get_unique_id().encode("utf-8"))
        md5.update(str(self.globals_.get_unique_id()).encode("utf-8"))
        return md5.hexdigest()


class MatrixValuedFunctionWrapper(_FunctionWrapperBase):
    def __init__(self, function, globals_, visitor, expose_to_python):
        assert set(function.arguments) & set(globals_.variables) == set(())
        _FunctionWrapperBase.__init__(
            self, function, globals_, visitor, expose_to_python
        )

        expression = function.args[1]
        assert isinstance(expression, ImmutableDenseMatrix)

        self.expressions = expression[:]

        self.variables = function.arguments
        self.globals_ = globals_

        self.m = expression.rows
        self.n = expression.cols

    def setup_code_generation(self):
        super().setup_code_generation()

        if 0:
            self._args = self.args
            if self.args:
                self.args = "{}, double _result[{}]".format(self.args, self.n * self.m)
            else:
                self.args = "double _result[{}]".format(self.n * self.m)

    @concat_generator_results
    def c_code(self, header_file_path):
        arrays_to_declare = set()
        scalars_to_declare = set()
        for alias in self.aliases:
            for symbol in alias.symbols:
                if isinstance(symbol, VectorElement):
                    arrays_to_declare.add(symbol.parent)
                else:
                    scalars_to_declare.add(symbol)

        declare_arrays = "\n|    ".join(
            "double {};".format(array) for array in arrays_to_declare
        )

        declare_scalars = "\n|    ".join(
            "double {};".format(atomic) for atomic in scalars_to_declare
        )

        _set_aliases = []

        for alias in self.aliases:
            for i, symbol in enumerate(alias.symbols):
                _set_aliases.append("{} = {}[{}];".format(symbol, alias.name, i))

        set_aliases = "\n|    ".join(_set_aliases)

        set_common_vars = "\n|    ".join(
            "double {} = {};".format(var, expr)
            for (var, expr) in self.common_expressions
        )

        set_results = "\n|    ".join(
            "__result[{}] = {};".format(i, expr)
            for (i, expr) in enumerate(self.simplified_expressions)
            if expr != "0"
        )

        yield align(
            """
        |extern "C" void _{name}({args}) {{
        |    {declare_arrays}
        |    {declare_scalars}
        |    {set_aliases}
        |    {set_common_vars}
        |    {set_results}
        |}}
        """.format(
                name=self.name,
                args=self.c_args_decl,
                declare_arrays=declare_arrays,
                declare_scalars=declare_scalars,
                set_aliases=set_aliases,
                set_results=set_results,
                set_common_vars=set_common_vars,
            )
        )

    @concat_generator_results
    def c_header(self):
        # we insert "_" at beginning of c functions to avoid clash with
        # cython functions:
        yield align(
            """
        |extern "C" void _{name}({args});
        """
        ).format(name=self.name, args=self.c_args_decl)

    @concat_generator_results
    def cython_code(self, header_file_path):
        if not self.expose_to_python:
            return
        # we insert "_" at beginning of c functions to avoid clash with
        # cython functions:
        for code in self.cython_header(header_file_path):
            yield code
        for code in self.cython_function_wrapper():
            yield code
        for code in self.cython_ufunc_wrapper():
            yield code

    def cython_header(self, header_file_path):
        yield align(
            """
        |cdef extern from "{header_file_path}":
        |    double _{name}({args})
        """.lstrip().format(
                header_file_path=header_file_path, name=self.name, args=self.c_args_decl
            )
        )

    def cython_function_wrapper(self):
        values = ", ".join(str(v.as_function_arg()) for v in self.in_variables)
        if values:
            c_args = values + ", &result[0]"
        else:
            c_args = "&result[0]"
        signature = _signature(self.expressions)
        yield align(
            """
        |cdef api int _api_{name}({args}):
        |    cdef np.ndarray[np.double_t, ndim=1] result
        |    cdef const char * _error_message
        |    cdef const char * _warning_message
        |
        |    clear_last_error_message()
        |    clear_last_warning_message()
        |
        |    result = np.zeros(({n} * {m},), dtype=np.double)
        |    _{name}({c_args})

        |    _error_message = c_get_last_error_message()
        |    if _error_message[0] != 0:
        |          return 1
        |    _warning_message = c_get_last_warning_message()
        |    if _warning_message[0] != 0:
        |          return 2
        |    return 0
        |
        """.lstrip().format(
                name=self.name,
                args=self.c_args_decl,
                c_args=c_args,
                m=self.m,
                n=self.n,
            )
        )
        yield align(
            """
        |def {name}({args}):
        |    cdef np.ndarray[np.double_t, ndim=1] result
        |    cdef const char * _error_message
        |    cdef const char * _warning_message
        |
        |    clear_last_error_message()
        |    clear_last_warning_message()
        |
        |    result = np.zeros(({n} * {m},), dtype=np.double)
        |    _{name}({c_args})
        |
        |    _warning_message = c_get_last_warning_message()
        |    if _warning_message[0] != 0:
        |        warnings.warn(str(_warning_message, encoding="utf-8"))
        |    _error_message = c_get_last_error_message()
        |    if _error_message[0] != 0:
        |          raise ArithmeticError(str(_error_message, encoding="utf-8") +
        |                " when evaluating {signature}"
        |                )
        |
        |    return result.reshape({m}, {n})
        """.lstrip().format(
                name=self.name,
                args=self.cython_args,
                c_args=c_args,
                signature=signature,
                m=self.m,
                n=self.n,
            )
        )

    def cython_ufunc_wrapper(self):
        yield align(
            """
            |def {name}_ufunc():
            |    raise RuntimeError("no ufunc created vector valued function {name}")
            |
            """
        ).format(name=self.name)


class VectorValuedFunctionWrapper(MatrixValuedFunctionWrapper):
    def __init__(self, function, globals_, visitor, expose_to_python):
        assert set(function.arguments) & set(globals_.variables) == set(())
        _FunctionWrapperBase.__init__(
            self, function, globals_, visitor, expose_to_python
        )

        self.n = len(self.expressions)

    def setup_code_generation(self):
        _FunctionWrapperBase.setup_code_generation(self)

    def cython_function_wrapper(self):
        args = [str(v.as_function_arg()) for v in self.in_variables]
        values = ", ".join(args + ["__result"])

        numba_signature = f"float64[:]({self.numba_type_decl})"
        vector_flags = [v.is_vector for v in self.in_variables]

        yield align(
            f"""
                |_api["{self.name}"] = ((ctypes.POINTER(ctypes.c_double), ({self.n},)),
                |                       [{self.ctypes_args}],
                |                       {args}, {vector_flags}, "{numba_signature}")
                """
        )

        yield align(
            """
        |cdef api int _api_{name}({args}):
        |    # cdef np.ndarray[np.double_t, ndim=1] result
        |    cdef const char * _error_message
        |    cdef const char * _warning_message
        |
        |    clear_last_error_message()
        |    clear_last_warning_message()
        |
        |    _{name}({values})

        |    _error_message = c_get_last_error_message()
        |    if _error_message[0] != 0:
        |          return 1
        |    _warning_message = c_get_last_warning_message()
        |    if _warning_message[0] != 0:
        |          return 2
        |    return 0
        """.lstrip().format(
                name=self.name,
                args=self.c_args_decl,
                values=values,
            )
        )


class RealValuedFunctionWrapper(_FunctionWrapperBase):
    def setup_code_generation(self):
        super().setup_code_generation()
        self.c_expression = self.simplified_expressions[0]

    @concat_generator_results
    def c_code(self, header_file_path):
        arrays_to_declare = set()
        scalars_to_declare = set()

        for alias in self.aliases:
            for symbol in alias.symbols:
                if isinstance(symbol, VectorElement):
                    arrays_to_declare.add(symbol.parent)
                else:
                    scalars_to_declare.add(symbol)

        declare_arrays = "\n|    ".join(
            "double {};".format(array) for array in arrays_to_declare
        )

        declare_scalars = "\n|    ".join(
            "double {};".format(atomic) for atomic in scalars_to_declare
        )

        _set_aliases = []

        for alias in self.aliases:
            for i, symbol in enumerate(alias.symbols):
                _set_aliases.append("{} = {}[{}];".format(symbol, alias.name, i))

        set_aliases = "\n|    ".join(_set_aliases)

        set_common_vars = "\n|    ".join(
            "double {} = {};".format(var, expr)
            for (var, expr) in self.common_expressions
        )

        # we insert "_" at beginning of c functions to avoid clash with
        # created python wrappers:
        result = align(
            """
        |extern "C" void _{name}({args}) {{
        |    {declare_arrays}
        |    {declare_scalars}
        |    {set_aliases}
        |    {set_common_vars}
        |    *__result = {c_expression};
        |
        |}}
        """.format(
                name=self.name,
                args=self.c_args_decl,
                declare_arrays=declare_arrays,
                declare_scalars=declare_scalars,
                set_aliases=set_aliases,
                set_common_vars=set_common_vars,
                c_expression=self.c_expression,
            )
        )
        yield result

    @concat_generator_results
    def c_header(self):
        # we insert "_" at beginning of c functions to avoid clash with
        # cython functions:
        yield align(
            """
        |extern "C" void _{name}({args});
        """
        ).format(name=self.name, args=self.c_args_decl)

    @concat_generator_results
    def cython_code(self, header_file_path):
        # we insert "_" at beginning of c functions to avoid clash with
        # cython functions:
        if not self.expose_to_python:
            return

        for code in self.cython_header(header_file_path):
            yield code
        for code in self.cython_function_wrapper():
            yield code

        for code in self.cython_ufunc_wrapper():
            yield code

    def cython_header(self, header_file_path):
        yield align(
            """
        |cdef extern from "{header_file_path}":
        |    double _{name}({args})
        """.lstrip().format(
                header_file_path=header_file_path, name=self.name, args=self.c_args_decl
            )
        )

    def cython_function_wrapper(self):
        args = [str(v.as_function_arg()) for v in self.in_variables]
        values = ", ".join(args + ["__result"])
        _checks = []

        for alias in self.aliases:
            msg = "{} does not have length {}".format(alias.name, len(alias.symbols))
            _checks.append(
                "assert len({}) == {}, '{}'".format(alias.name, len(alias.symbols), msg)
            )

        numba_signature = f"float64({self.numba_type_decl})"

        vector_flags = [v.is_vector for v in self.in_variables]

        yield align(
            f"""
                |_api["{self.name}"] = ((ctypes.POINTER(ctypes.c_double), None),
                |                       [{self.ctypes_args}],
                |                       {args}, {vector_flags}, "{numba_signature}")
                """
        )

        yield align(
            """
        |cdef api int _api_{name}({args}):
        |
        |    cdef const char * _error_message
        |    cdef const char * _warning_message
        |
        |    clear_last_error_message()
        |    clear_last_warning_message()
        |
        |    _{name}({values})
        |
        |    _error_message = c_get_last_error_message()
        |    if _error_message[0] != 0:
        |          return 1
        |    _warning_message = c_get_last_warning_message()
        |    if _warning_message[0] != 0:
        |          return 2
        |    return 0
        |
        """.lstrip().format(
                name=self.name,
                args=self.c_args_decl,
                values=values,
            )
        )

    def cython_ufunc_wrapper(self):
        out_dim = len(self.variables)
        if out_dim == 0:
            return

        sizes_check = (
            "|    if len(set(({},)) - set((1,))) > 1:\n"
            "|        raise ValueError('lengths must be one or agree.')"
        ).format(", ".join(f"{vi._name()}.shape[0]" for vi in self.in_variables))

        out_size_computation = ("|    cdef size_t __out_size = max({}, 0)").format(
            ", ".join(f"{vi._name()}.shape[0]" for vi in self.in_variables)
        )

        values_decl = "\n|    ".join(
            "cdef double{} __v{}".format(
                "[:]" if self.in_variables[i].is_vector else "", i
            )
            for i in range(out_dim)
        )

        set_values = "\n|        ".join(
            f"__v{j} = {arg._name()}[0] if {arg}.shape[0] == 1 else {arg._name()}[__i]"
            for (j, arg) in enumerate(self.in_variables)
        )

        values = ", ".join(
            [
                f"&__v{j}[0]" if var.is_vector else f"__v{j}"
                for j, var in enumerate(self.in_variables)
            ]
            + ["&__result_view[__i]"]
        )

        signature = _signature(self.expressions[0])

        yield align(
            """
        |@cython.boundscheck(False)
        |@cython.wraparound(False)
        |def {name}_ufunc({ufunc_args}):
        |    cdef np.ndarray[np.double_t, ndim=1, cast=True] __result
        |    cdef str __error_message, __warning_message
        |    cdef size_t __i
        {sizes_check}
        {out_size_computation}
        |    {values_decl}
        |    __result = np.empty((__out_size,), dtype=np.double)
        |    cdef double[:] __result_view = __result
        |    clear_last_warning_message()
        |    clear_last_error_message()
        |
        |    for __i in range(__out_size):
        |        {set_values}
        |        _{name}({values})
        |
        |    __warning_message = get_last_warning_message()
        |    if __warning_message:
        |        warnings.warn(str(__warning_message, encoding="utf-8"))
        |
        |    __error_message = get_last_error_message()
        |    if __error_message:
        |          raise ArithmeticError(str(__error_message, encoding="utf-8") +
        |                " when evaluating {signature}"
        |                )
        |    return __result
        """.lstrip().format(
                name=self.name,
                ufunc_args=self.ufunc_args,
                out_dim=out_dim,
                sizes_check=sizes_check,
                out_size_computation=out_size_computation,
                values_decl=values_decl,
                set_values=set_values,
                values=values,
                signature=signature,
            )
        )



# This file is part of sympy2c.
#
# Copyright (C) 2013-2022 ETH Zurich, Institute for Particle and Astrophysics and SIS
# ID.
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.

from functools import partial

from sympy import Symbol

"""
patch existing sympy Symbol class by adding some special methods we
need for code generation.
"""


def c_args_decl(self):
    return self


Symbol.c_args_decl = c_args_decl
Symbol.as_function_arg = c_args_decl


def cython_decl(self):
    return "double " + str(self)


Symbol.cython_decl = cython_decl


def c_decl(self):
    return "double " + str(self)


Symbol.c_decl = c_decl


def ctypes_decl(self):
    return "ctypes.c_double"


Symbol.ctypes_decl = ctypes_decl


def numba_type_decl(self):
    return "float64"


Symbol.numba_type_decl = numba_type_decl


def _name(self):
    return self.name


Symbol._name = _name

Symbol.is_vector = False


def setter_c_code(self):
    return """extern "C" void _set_{name}(double _v){{ {name} = _v; }}""".format(
        name=self.name
    )


Symbol.setter_c_code = setter_c_code


def getter_c_code(self):
    return """extern "C" double _get_{name}(){{ return {name}; }}""".format(
        name=self.name
    )


Symbol.getter_c_code = getter_c_code


def setter_header_code(self):
    return "void _set_{name}(double _v); ".format(name=self.name)


Symbol.setter_header_code = setter_header_code


def getter_header_code(self):
    return "double _get_{name}(); ".format(name=self.name)


Symbol.getter_header_code = getter_header_code

Symbol.setter_cython_header_code = setter_header_code
Symbol.getter_cython_header_code = getter_header_code


def cython_globals_setter_code(self):
    from .utils import align

    return (
        align(
            """
    |if "{name}" in _g:
    |    if _g["{name}"] is None:
    |       _set_{name}(np.nan)
    |    else:
    |       _set_{name}(<double>_g["{name}"])
    """
        )
        .format(name=self.name)
        .rstrip()
    )


Symbol.cython_globals_setter_code = cython_globals_setter_code


def cython_globals_getter_code(self):
    return """result['{name}'] = _get_{name}()""".format(name=self.name)


Symbol.cython_globals_getter_code = cython_globals_getter_code

Symbol = partial(Symbol, real=True)


def symbols(txt):
    return [Symbol(n) for n in txt.split()]
